/**********************************************************/
/* hybridGM, version 1.0
/*
 * a modified version of gNovelty+ which calls march_ks on simplified
 * instances of the problem created with the help of  search space partitions
 */
/*

/************************************/
/* Standard includes                */
/************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "march.h"
#include "global.h"
#include "rtime.h"

/************************************/
/* Compilation flags                */
/************************************/
#define WEIGHTING 1

//#define DETERMINISTIC_SMOOTHING 1

/**********************************************************/
/* DLS specific variables                                 */
/**********************************************************/
#define CWTYPE int
//#define CWTYPE double

inline double max(double a, double b) {
	return (a > b) ? a : b;
}

inline double min(double a, double b) {
	return (a < b) ? a :b;
}

#define _EXP(X) max(0.001, 1.0 + (X)) /*Fake update that is faster than machine exponent.*/

//=== Weighting controlled variables
CWTYPE clauseWeight[MAXCLAUSE];		// weight of a clause
CWTYPE weightDiff[MAXATOM+1];		// the total weight difference if a variable if flipped
CWTYPE weightTotal;					// total clause weight

CWTYPE pseudoStepSize;				// initialised in main()
CWTYPE trace[MAXCLAUSE];
CWTYPE traceDiscount;
CWTYPE stepSize[MAXCLAUSE];

int numWeight;						// number of clauses with weight > 1
int weightedClause[MAXCLAUSE];		// clauses which have weight > 1:
int whereWeight[MAXCLAUSE];		// same as for whereFalse:
int smoothStages, weightCounter;

int smoothProb = 100;

//=== AdaptNovelty+ controlled variables
int varLastChange[MAXATOM+1];
int lastAdaptFlip;
int lastAdaptNumFalse;

int adaptFlag = 1;

#define _weight_getScore(VAR)		(weightDiff[VAR])
#define _weight_setScore(VAR, VAL)	(weightDiff[VAR] = (CWTYPE) VAL)
#define _weight_adjScore(VAR, VAL)	(weightDiff[VAR] += VAL)
#define _weight_incScore(VAR, CLS)	(weightDiff[VAR] += clauseWeight[CLS])
#define _weight_decScore(VAR, CLS)	(weightDiff[VAR] -= clauseWeight[CLS])

#define _num_getScore(VAR)		(numDiff[VAR])
#define _num_setScore(VAR, VAL)	(numDiff[VAR] = VAL)
#define _num_adjScore(VAR, VAL)	(numDiff[VAR] += VAL)
#define _num_incScore(VAR, CLS)	(numDiff[VAR] += 1)
#define _num_decScore(VAR, CLS)	(numDiff[VAR] -= 1)

#define setScore(VAR, VAL) {		\
	_weight_setScore(VAR, VAL);		\
	_num_setScore(VAR, VAL);		\
}

#define adjScore(VAR, VAL) {		\
	_weight_adjScore(VAR, VAL);		\
	_num_adjScore(VAR, VAL);		\
}

#define incScore(VAR, CLS) {		\
	_weight_incScore(VAR, CLS);		\
	_num_incScore(VAR, CLS);		\
}

#define decScore(VAR, CLS) {		\
	_weight_decScore(VAR, CLS);		\
	_num_decScore(VAR, CLS);		\
}



/**********************************************************/
/* ANOV parameters                                        */
/**********************************************************/
#define PROBABILITY int

PROBABILITY walkProb = 1;	// default random walk probability
PROBABILITY noise =50;		// default noise (in percent)

int invPhi = 5;
int invTheta = 6;

/**********************************************************/
/*   Forward declarations                                 */
/**********************************************************/
void parseParameters(int argc,char *argv[]);
void parseParametersComp(int argc,char *argv[]);
void scanInt(int argc, char *argv[], int i, int *varptr);
void scanFloat(int argc, char *argv[], int i, float *varptr);
void printHeader();
void printStatsHeader();
int  compareAbsInts(const void *a, const void *b);
void initInstance();
void init();
int  pickVarG2WSAT();
int  pickVarANov();
void flipAtomG2WSAT(int toflip);
void flipAtomANov(int toflip);
void updateClauseWeights_Linear();
void updateClauseWeights_NULL();
void smooth();
void adaptNoveltyNoise();
void updateStatsEndFlip();
void updateStatsEndTry();
void printStatsEndTry();
CWTYPE computeCurrentCost();
void printFinalStats();
void printFinalStatsComp();
void saveSolution();
void printSolutionG();

int *neighbourVar[MAXATOM+1];

int candVar[MAXATOM+1];
int isCandVar[MAXATOM+1];
int numCandVar;

int flags;
int numWalks, numDBests, numFreqs, numSBests;

int minClauseSize, maxClauseSize;
void (*updateClauseWeights)() = NULL;
void (*flipAtom)(int toflip) = NULL;
int (*pickVar)()= NULL;
FILE *fpInput;


void printStatsEndTryM();

//___________________________________________________---------------------_________________________________
int callMarchKS(int* assignment, int freeAtoms, int** resClause, int marchResult[]);

int hybrid=1;
int barrier=1;
int freeDegree;
int freeDegreeStep;
int regExp[MAXATOM+1];
int oldRegExp[MAXATOM+1];
int maxRegExp[MAXATOM+1];
int lastCallMarch=0;
int intervalCallMarch;
int marchCount=0;

//=== Ringbuf
#define RINGBUFSIZE 5*MAXATOM
int ringBuf[RINGBUFSIZE];
int ringBufPointer=0;
int ringBufReader=0;
int ringBufValid=0;
void initRingBuf();
void initReadRingBuf();
int readRingBuf();
void insertRingBuf(int var);

double starttime;
//___________________________________________________---------------------_________________________________


/************************************/
/* Main                             */
/************************************/
int main(int argc,char *argv[]) {
	int try, var;
	seed = genRandomSeed();
	parseParametersComp(argc, argv);
	//parseParameters(argc, argv);
	srandom(seed);
	initInstance();
	initRingBuf();

	if (maxClauseSize <= 3)  {
		printf("c Heuristic: AdaptG2WSAT \n");
		printf("c Hybrid barrier : 1 \n");
		flipAtom= &flipAtomG2WSAT;
		pickVar= &pickVarG2WSAT;
		updateClauseWeights = &updateClauseWeights_Linear;
		smoothProb = 33;
		adaptFlag = 1;		noise = 0;
		hybrid=1;
		freeDegree=numAtoms/2;
	}
//	else {
//		printf("c Heuristic : AdaptG2WSAT\n");
//		flipAtom= &flipAtomG2WSAT;
//		pickVar= &pickVarG2WSAT;
//		updateClauseWeights = &updateClauseWeights_NULL;
//		adaptFlag = 1;		noise = 0;
//		printf("c Hybrid barrier : 0 \n");
//		hybrid=0;
//		//freeDegree=numAtoms*9/20;
//		}

	else {
		if (numClauses>10000){
			printf("c Heuristic : ANov\n");
			flipAtom= &flipAtomANov;
			pickVar= &pickVarANov;
		}
		else{
			printf("c Heuristic : G2WSAT\n");
			flipAtom= &flipAtomG2WSAT;
			pickVar= &pickVarG2WSAT;
		}
		updateClauseWeights = &updateClauseWeights_NULL;
		adaptFlag = 1;		noise = 0;
		printf("c Hybrid barrier : 1 \n");
		hybrid=1;
		freeDegree=numAtoms*9/20;
	}



	int marchAnswer=0;
	int i,collecting=0,freeDegCount=0,counter=0,counter1=0,j,diffRegs=0,diffValues=0,maxDiffValue=0;
	int collectingBackwards=1;
	int diffToMax=0,posMax;
	int startFlip;
	int fCollected=0;//Flips collected so far in this collection period
	int numFalseB,totalNumFalse=0; //number of false clauses in buffer
	int theFalseC; //the clause that ist unsatisfied in the local minima
	int marchResult[numAtoms+1];
	//freeDegree=numAtoms/2;
	freeDegreeStep= numAtoms/20;
	int blockedIntervall=numClauses*2;


	totalFlips = totalNullFlips = 0;
	totalSuccessFlips = totalSuccessNullFlips = 0;
	numSuccessTries = 0;

	printStatsHeader();
	initTimeCounter();

	lowCost = BIG;
	for (try = 1; try <= numTries; try ++) {

		starttime = elapsedTime();
		flip = nullFlip = 0;
		pseudoStepSize = pow(10, -15);		// required for the stepsize weighting
		traceDiscount = 0.99;

		init();
		while (numFalse > targetCost){
			//while ((numFalse > targetCost) && (flip - nullFlip < cutoff)) {
			flip++;
			var = pickVar();
			flipAtom(var);
			varLastChange[var] = flip;

			if (adaptFlag && flags)
				adaptNoveltyNoise();

			updateStatsEndFlip();

			//--------------------Ringbuffer-------Construct RegEXP & Call March------------//
			if (hybrid==1){
				insertRingBuf(var);
				if (collecting==1){
					if (freeDegree>freeDegCount){
						if (regExp[var]!=2){
							freeDegCount++;
							regExp[var]=2;
						}
						if (collectingBackwards==1){
							i=readRingBuf();
							if (i==-1){
								//printf("c Buffer BABUT: Aborting collecting backwards\n");
								collectingBackwards=0;
							}
							else
								if (regExp[i]!=2){
									freeDegCount++;
									regExp[i]=2;
								}
						}
					}
					else{
						collecting=0;
						counter++;
						if (counter==5000) hybrid=0;//avoid MemoryOut
						int k;
					//	printf("%4d Barrier: %d Calling March with %d variales from FLIP: (%ld -%ld ) = %ld",counter,barrier,freeDegCount,startFlip,flip,flip-startFlip);
						marchAnswer=callMarchKS(regExp, freeDegCount, resClause, marchResult);
//						printf(" DONE__\n ");
						lastCallMarch=flip;

						if(marchAnswer==20) {
							marchCount++;
							//printf("COMPUTED \n");
						}
						if (marchAnswer==21){
							marchCount++;
							//printf("Unary \n");
							freeDegree+=freeDegreeStep;
						}
						else if (marchAnswer==10){
							//printf("\n");
							for (j=1;j<=numAtoms;j++)
								atom[j]=marchResult[j];
							numFalse=0;
							updateStatsEndTry();
							expertime = elapsedTime();
							//printStatsEndTryM();
							printf("c March( %d )\n",counter);
							fflush(stdout);
							printFinalStatsComp();

							//						printf("randWalks: %4.2f\n decreasingSteps: %4.2f\n freq: %10d\n fb: %4.2f\n secondBest: %4.2f\n",
							//								(float)numWalks/(float)flip, (float)numDBests/(float)flip, numFreqs, (float)(flip-numWalks-numDBests-numFreqs-numSBests)/(float)flip, (float)numSBests/(float)flip);


						}

					}
				}
				if ((numFalse<=barrier)&&(collecting!=1)&&(flip-lastCallMarch>blockedIntervall)){//&&(flip-lastCallMarch>blockedIntervall)
					collecting=1;
					collectingBackwards=1;
					startFlip=flip;
					freeDegCount=0;

					for (i=1;i<=numAtoms;i++){
						regExp[i]=atom[i];
					}
					initReadRingBuf();
				}
			}
		}
		//trytime = elapsedTime() - starttime;
		updateStatsEndTry();
		expertime = elapsedTime();
		//printStatsEndTry();

		//		printf("randWalks: %4.2f\n decreasingSteps: %4.2f\n freq: %10d\n fb: %4.2f\n secondBest: %4.2f\n",
		//										(float)numWalks/(float)flip, (float)numDBests/(float)flip, numFreqs, (float)(flip-numWalks-numDBests-numFreqs-numSBests)/(float)flip, (float)numSBests/(float)flip);

	}
	//expertime = elapsedTime();
	//	printFinalStats();
	printf("c gNovelty+ (M  %d )\n",counter);	fflush(stdout);
	printFinalStatsComp();
	return 0;
}




int pickVarG2WSAT(void) {
	CWTYPE score, bestValue, secondBestValue;
	int bestVar=0, secondBestVar, youngest, lastChange;
	register int i, j, clause, var;
	int *clsptr;
	flags = TRUE;

	if (numCandVar > 0) {	//=== greedy step
		bestValue = BIG;  //CWBIG;	//BIG;
		lastChange = flip;

		i = NOVALUE;
		clsptr = candVar;
		for (j = 0; j < numCandVar; j++, clsptr++) {
			var = *clsptr;
			score = _weight_getScore(var);
			if (score < 0) {
				if (score < bestValue) {
					bestValue = score;
					lastChange = varLastChange[var];
					bestVar = var;
				} else
					if (score == bestValue) {
						if (varLastChange[var] < lastChange) {
							lastChange = varLastChange[var];
							bestVar = var;
						}
					}
			}
			else {
				i = j;		isCandVar[var] = FALSE;
				break;
			}
		}
		clsptr++;

		if (i != NOVALUE) {
			for (j = i+1; j < numCandVar; j++, clsptr++) {
				var = *clsptr;
				score = _weight_getScore(var);
				if (score < 0) {
					candVar[i++] = var;
					if (score < bestValue) {
						bestValue = score;
						lastChange = varLastChange[var];
						bestVar = var;
					} else
						if (score == bestValue) {
							if (varLastChange[var] < lastChange) {
								lastChange = varLastChange[var];
								bestVar = var;
							}
						}
				} else {
					isCandVar[var] = FALSE;
				}
			}
			numCandVar = i;
		}
	}



	if (random()%100 < walkProb) {	//=== Normal AdaptNovelty+
		clause = falseClause[random() % numFalse];
		bestVar = Var(clause,(random() % size[clause]));
		//numWalks++;

	}
	else
		if (numCandVar > 0) {
			flags = FALSE;
//			numDBests++;
			//printf("oOo"); fflush(stdout);
		}
		else {

#ifdef WEIGHTING
			(*updateClauseWeights)();
#endif

			bestValue = secondBestValue = BIG;
			bestVar = secondBestVar = NOVALUE;

			clause = falseClause[random() % numFalse];
			clsptr = resClause[clause];
			youngest = varLastChange[abs(*clsptr)];

			for (j = 0; j < size[clause]; j++) {
				var = abs(*clsptr);

				//score = break[var] - make[var];
				//score = _num_getScore(var);
				score = _weight_getScore(var);
				lastChange = varLastChange[var];

				if (lastChange > youngest) {
					youngest = varLastChange[var];
				}

				if ( (score < bestValue) || ((score == bestValue) && (lastChange < varLastChange[bestVar])) ) {
					secondBestVar = bestVar;
					secondBestValue = bestValue;

					bestVar = var;
					bestValue = score;
				} else if ( (score < secondBestValue) || ((score == secondBestValue) && (lastChange < varLastChange[secondBestVar])) ) {
					secondBestVar = var;
					secondBestValue = score;
				}

				clsptr++;
			}

			/*
		if ( (varLastChange[bestVar] == youngest) && (random()%100 < noise) ) {
			bestVar = secondBestVar;
			//numSBests++;
		}
			 */

			if (varLastChange[bestVar] == youngest) {
				if ( (_num_getScore(bestVar) == 0) || (random()%100 < noise) ) {
					bestVar = secondBestVar;
				//	numSBests++;
				}
			}

		}
	return bestVar;
}

int pickVarANov(void) {
	CWTYPE score, bestValue, secondBestValue;
	int bestVar=0, secondBestVar, youngest, lastChange;
	register int i, j, clause, var;
	int *clsptr;

	flags = TRUE;
	//TODO : umdrehen nach vorne ziehen

	if (random()%100 < walkProb) {	//=== Normal AdaptNovelty+
		clause = falseClause[random() % numFalse];
		bestVar = Var(clause,(random() % size[clause]));
		numWalks++;

	}
	else{

		bestValue = secondBestValue = BIG;
		bestVar = secondBestVar = NOVALUE;

		clause = falseClause[random() % numFalse];
		clsptr = resClause[clause];
		youngest = varLastChange[abs(*clsptr)];

		for (j = 0; j < size[clause]; j++) {
			var = abs(*clsptr);

			//score = break[var] - make[var];
			//score = _num_getScore(var);
			score = _weight_getScore(var);
			lastChange = varLastChange[var];

			if (lastChange > youngest) {
				youngest = varLastChange[var];
			}

			if ( (score < bestValue) || ((score == bestValue) && (lastChange < varLastChange[bestVar])) ) {
				secondBestVar = bestVar;
				secondBestValue = bestValue;

				bestVar = var;
				bestValue = score;
			} else if ( (score < secondBestValue) || ((score == secondBestValue) && (lastChange < varLastChange[secondBestVar])) ) {
				secondBestVar = var;
				secondBestValue = score;
			}

			clsptr++;
		}

		/*
		if ( (varLastChange[bestVar] == youngest) && (random()%100 < noise) ) {
			bestVar = secondBestVar;
			//numSBests++;
		}
		 */

		if (varLastChange[bestVar] == youngest) {
			if ( (_num_getScore(bestVar) == 0) || (random()%100 < noise) ) {
				bestVar = secondBestVar;
				numSBests++;
			}
		}
	}

	return bestVar;
}



void updateClauseWeights_NULL() {
}

void updateClauseWeights_Linear() {
	register int c, i, j, var;

	for (i = 0; i < numFalse; i++) {
		c = falseClause[i];

		if (++clauseWeight[c] == 2) {
			//=== Add clause to weighted list
			whereWeight[c] = numWeight;
			weightedClause[numWeight] = c;
			++numWeight;
		}
		for (j = 0; j < size[c]; j++) {
			var = Var(c, j);
			_weight_adjScore(var, -1);
			if (!isCandVar[var] && (_weight_getScore(var) < 0) && (varLastChange[var] < flip-1)) {
				candVar[numCandVar++] = var;
				isCandVar[var] = TRUE;
			}
		}
	}

#ifdef DETERMINISTIC_SMOOTHING
	if(++weightCounter > smoothProb) {
		smooth();
		weightCounter = 0;
	}
#else
	if (random()%1000 < smoothProb*10) smooth();
#endif

}

void updateClauseWeights_Exp() {
	register int c, i, j, var;

	for (i = 0; i < numFalse; i++) {
		c = falseClause[i];

		stepSize[c] = stepSize[c] * exp(- 1.0 * pseudoStepSize *  trace[c]);
		trace[c] = traceDiscount * trace[c] + _EXP(stepSize[c]);

		clauseWeight[c] += stepSize[c];

		//		fprintf(stderr, "  oOo %d: %E -- %E -- %E oOo  ", c, stepSize[c], trace[c], clauseWeight[c]);

		//for (j = 0; j < size[c]; j++) _weight_adjScore(abs(resClause[c][j]), -stepSize[c]);

		for (j = 0; j < size[c]; j++) {
			var = Var(c, j);
			_weight_adjScore(var, -stepSize[c]);
			if (!isCandVar[var] && (_weight_getScore(var) < 0.0) && (varLastChange[var] < flip-1)) {
				candVar[numCandVar++] = var;
				isCandVar[var] = TRUE;
			}
		}
	}
}

void smooth() {
	register int c, i, j;

	int startNumWeight = numWeight;

	for (i = 0; i < startNumWeight; i++) {
		c = weightedClause[i];
		if (--clauseWeight[c] == 1) {
			--numWeight;
			weightedClause[whereWeight[c]] = weightedClause[numWeight];
			whereWeight[weightedClause[numWeight]] = whereWeight[c];
		}
		if (numTrueLit[c] == 0) {
			for (j = 0; j < size[c]; j++) _weight_adjScore(Var(c,j), 1);
		}
		else if (numTrueLit[c] == 1) _weight_adjScore(critVar[c], -1);
	}
	++smoothStages;
}

void adaptNoveltyNoise() {
	if ( (flip - lastAdaptFlip) > (numClauses / invTheta) ) {
		noise += (int) ((100 - noise) / invPhi);
		lastAdaptFlip = flip;
		lastAdaptNumFalse = numFalse;
		//printf("Noise++\n");
	} else if (numFalse < lastAdaptNumFalse) {
		noise -= (int) (noise / invPhi / 2);
		lastAdaptFlip = flip;
		lastAdaptNumFalse = numFalse;
		//printf("Noise--\n");
	}
}

void init() {
	register int i, j, lit;

	for (i = 0; i < numClauses; i++) {
		numTrueLit[i] = 0;
		clauseWeight[i] = 1;
		//
		trace[i] = 0.0;
		stepSize[i] = pseudoStepSize;
	}

	numFalse = 0;
	numWeight = 0;
	nullFlip = 0;
	smoothStages = weightCounter = 0;

	for (i = 1; i < numAtoms+1; i++) {
		setScore(i, 0);
		varLastChange[i] = 0;
	}

	for (i = 1; i < numAtoms+1; i++) atom[i] = random()%2;

	//=== Initialize the gradient cost
	for (i = 0; i < numClauses; i++) {
		for (j = 0; j < size[i]; j++){
			lit = Lit(i,j);
			if ((lit > 0) == atom[abs(lit)]) {
				numTrueLit[i]++;
				critVar[i] = abs(lit);
			}
		}
		if (numTrueLit[i] == 0) {
			whereFalse[i] = numFalse;
			falseClause[numFalse] = i;
			numFalse++;

			//=== these variables make clause i
			for (j = 0; j < size[i]; j++) {
				decScore(Var(i,j), i);
			}
		} else if (numTrueLit[i] == 1) {
			//=== this variable breaks clause i
			incScore(critVar[i], i);
		}
	}
	lowCost = numFalse;

	numCandVar = 0;
	for (i = 1; i < numAtoms+1; i++) {
		if (_weight_getScore(i) < 0) {
			candVar[numCandVar++] = i;
			isCandVar[i] = TRUE;
		} else {
			isCandVar[i] = FALSE;
		}
	}

	// Adapt Novelty+ Noise
	if (adaptFlag) {
		lastAdaptFlip = flip;
		lastAdaptNumFalse = numFalse;
		noise = 0;
	}
}

void flipAtomG2WSAT(int toflip) {
	register int i, j, c, var;
	int toEnforce;		//=== literal to enforce
	int numocc;
	int *occptr, *ptr;
	
		if (toflip == NOVALUE) {
			totalNullFlips++;
			nullFlip++;
			return;
		}

	//	CWTYPE oldScore = getScore(toflip);

	if (atom[toflip] > 0) toEnforce = -toflip;
	else toEnforce = toflip;

	//=== Flip the variable here.
	atom[toflip] = 1-atom[toflip];

	for (ptr = neighbourVar[toflip]; *ptr != NOVALUE; ptr++) {
		i = *ptr;
		if (_weight_getScore(i) >= 0) isCandVar[i] = FALSE;
		else isCandVar[i] = TRUE;
	}

	//=== Examine all occurrences of literals with toEnforce with old sign.
	//=== (these literals are made false; numTrueLit must be decremented in that clause)
	numocc = numOccurence[MAXATOM-toEnforce];
	occptr = occurence[MAXATOM-toEnforce];
	for (i = 0; i < numocc ;i++) {
		c = *(occptr++);
		if (--numTrueLit[c] == 0) {
			// clause c is no longer satisfied by lit with toflip; it's now false.
			falseClause[numFalse] = c;
			whereFalse[c] = numFalse++;

			//=== Decrement toflip's break.
			decScore(toflip, c);
			//=== Increment make of all vars in clause c.
			for (j = 0; j < size[c]; j++) {
				decScore(Var(c,j), c);
				//if (Var(c,j) == toflip) printf("DEC %d== ", c);
				//fflush(stdout);
			}
		} else if (numTrueLit[c] == 1) {
			//=== Find the lit in this clause that makes it true, and inc its break.
			for (j = 0; j < size[c]; j++) {
				var = Var(c,j);
				if ((Lit(c,j) > 0) == atom[var]) {
					critVar[c] = var;
					incScore(var, c);
					break;
				}
			}
		}
	}

	//=== Examines all occurrences of literal with toEnforce with new sign
	//=== (these literals are made true; numTrueLit must be incremented in that clause)
	numocc = numOccurence[MAXATOM+toEnforce];
	occptr = occurence[MAXATOM+toEnforce];
	for (i = 0; i < numocc; i++) {
		c = *(occptr++);
		if (++numTrueLit[c] == 1){
			falseClause[whereFalse[c]] = falseClause[--numFalse];
			whereFalse[falseClause[numFalse]] = whereFalse[c];

			critVar[c] = toflip;
			incScore(toflip, c);
			//=== Decrement the make of all vars in clause c.
			for (j = 0; j < size[c]; j++) {
				incScore(Var(c,j), c);
				//if (Var(c,j) == toflip) printf("ASC %d== ", c);
				//fflush(stdout);
			}
		} else if (numTrueLit[c] == 2){
			//=== Find the lit in this clause other than toflip that makes it true,
			//=== and decrement its break.
			for (j=0; j<size[c]; j++){
				var = Var(c,j);
				if( ((Lit(c,j) > 0) == atom[var]) && (toflip != var) ) {
					decScore(var, c);
					break;
				}
			}
		}
	}

	/*
	if (oldScore != 0) {
		assert(oldScore != getScore(toflip));
	} else {
		printf("Zero cost move\n");
		fflush(stdout);
	}
	 */
	//TODO : diese sache scheint zu g2wsat zu gehören und aktualisiert nur die candVars

	for (ptr = neighbourVar[toflip]; *ptr != NOVALUE; ptr++) {
		i = *ptr;
		if (_weight_getScore(i) < 0) {
			if (!isCandVar[i]) {
				candVar[numCandVar++] = i;
				isCandVar[i] = TRUE;
			}
		}
	}
}

void flipAtomANov(int toflip) {
	register int i, j, c, var;
	int toEnforce;		//=== literal to enforce
	int numocc;
	int *occptr, *ptr;
	
		if (toflip == NOVALUE) {
			totalNullFlips++;
			nullFlip++;
			return;
		}

	//	CWTYPE oldScore = getScore(toflip);

	if (atom[toflip] > 0) toEnforce = -toflip;
	else toEnforce = toflip;

	//=== Flip the variable here.
	atom[toflip] = 1-atom[toflip];

	for (ptr = neighbourVar[toflip]; *ptr != NOVALUE; ptr++) {
		i = *ptr;
		if (_weight_getScore(i) >= 0) isCandVar[i] = FALSE;
		else isCandVar[i] = TRUE;
	}

	//=== Examine all occurrences of literals with toEnforce with old sign.
	//=== (these literals are made false; numTrueLit must be decremented in that clause)
	numocc = numOccurence[MAXATOM-toEnforce];
	occptr = occurence[MAXATOM-toEnforce];
	for (i = 0; i < numocc ;i++) {
		c = *(occptr++);
		if (--numTrueLit[c] == 0) {
			// clause c is no longer satisfied by lit with toflip; it's now false.
			falseClause[numFalse] = c;
			whereFalse[c] = numFalse++;

			//=== Decrement toflip's break.
			decScore(toflip, c);
			//=== Increment make of all vars in clause c.
			for (j = 0; j < size[c]; j++) {
				decScore(Var(c,j), c);
				//if (Var(c,j) == toflip) printf("DEC %d== ", c);
				//fflush(stdout);
			}
		} else if (numTrueLit[c] == 1) {
			//=== Find the lit in this clause that makes it true, and inc its break.
			for (j = 0; j < size[c]; j++) {
				var = Var(c,j);
				if ((Lit(c,j) > 0) == atom[var]) {
					critVar[c] = var;
					incScore(var, c);
					break;
				}
			}
		}
	}

	//=== Examines all occurrences of literal with toEnforce with new sign
	//=== (these literals are made true; numTrueLit must be incremented in that clause)
	numocc = numOccurence[MAXATOM+toEnforce];
	occptr = occurence[MAXATOM+toEnforce];
	for (i = 0; i < numocc; i++) {
		c = *(occptr++);
		if (++numTrueLit[c] == 1){
			falseClause[whereFalse[c]] = falseClause[--numFalse];
			whereFalse[falseClause[numFalse]] = whereFalse[c];

			critVar[c] = toflip;
			incScore(toflip, c);
			//=== Decrement the make of all vars in clause c.
			for (j = 0; j < size[c]; j++) {
				incScore(Var(c,j), c);
				//if (Var(c,j) == toflip) printf("ASC %d== ", c);
				//fflush(stdout);
			}
		} else if (numTrueLit[c] == 2){
			//=== Find the lit in this clause other than toflip that makes it true,
			//=== and decrement its break.
			for (j=0; j<size[c]; j++){
				var = Var(c,j);
				if( ((Lit(c,j) > 0) == atom[var]) && (toflip != var) ) {
					decScore(var, c);
					break;
				}
			}
		}
	}

	/*
	if (oldScore != 0) {
		assert(oldScore != getScore(toflip));
	} else {
		printf("Zero cost move\n");
		fflush(stdout);
	}
	 */
	//TODO : diese sache scheint zu g2wsat zu gehören und aktualisiert nur die candVars
	//is not needed for the AdaptNovlety part
	/*
	for (ptr = neighbourVar[toflip]; *ptr != NOVALUE; ptr++) {
		i = *ptr;
		if (_weight_getScore(i) < 0) {
			if (!isCandVar[i]) {
				candVar[numCandVar++] = i;
				isCandVar[i] = TRUE;
			}
		}
	}
	 */
}

void parseParameters(int argc,char *argv[]) {
	int i;

	numTries = DEFAULT_NUMTRIES;
	cutoff = DEFAULT_CUTOFF_FLIPS;
	targetCost = DEFAULT_TARGETCOST;
	printSol = DEFAULT_PRINTSOL;

	for (i = 1; i < argc-1; i++) {
		if (strcmp(argv[i],"-seed") == 0) scanInt(argc,argv,++i,&seed);
		else if (strcmp(argv[i],"-cutoff") == 0) scanInt(argc,argv,++i,&cutoff);
		else if (strcmp(argv[i],"-tries") == 0) scanInt(argc,argv,++i,&numTries);
		else if (strcmp(argv[i],"-targetcost") == 0) scanInt(argc,argv,++i,&targetCost);
		else if (strcmp(argv[i],"-noise") == 0) scanInt(argc,argv,++i,&noise);
		else if (strcmp(argv[i],"-walkprob") == 0) scanInt(argc,argv,++i,&walkProb);
		else if (strcmp(argv[i],"-smoothprob") == 0) scanInt(argc,argv,++i,&smoothProb);
		else if (strcmp(argv[i],"-printsol") == 0) printSol = TRUE;
		else {
			if (strcmp(argv[i],"-help")!=0 && strcmp(argv[i],"-h")!=0 )
				fprintf(stderr, "\nBad argument %s\n\n", argv[i]);
			fprintf(stderr, "General parameters:\n");
			fprintf(stderr, "  -seed N (integer, default: quasi-random based on system time)\n");
			fprintf(stderr, "  -cutoff N (integer, default: %d)\n", cutoff);
			fprintf(stderr, "  -tries N (integer, default: %d)\n", numTries);
			fprintf(stderr, "  -noise N (percent, default: %d)\n", noise);
			fprintf(stderr, "  -walkprob N (percent, default: %d)\n", walkProb);
			fprintf(stderr, "  -smoothprob N (percent, default: %d)\n", smoothProb);
			fprintf(stderr, "  -targetcost N = find assignments of cost <= N (MAXSAT) (integer, default: %d)\n", targetCost);
			fprintf(stderr, "  -printsol\n");
			fprintf(stderr, "\nUsage:\n");
			fprintf(stderr, "  sls -parameters < instance.cnf\n\n");
			exit(-1);
		}
	}
	fpInput = fopen(argv[argc-1], "r");
	if (fpInput == NULL) {
		fprintf(stderr, "ERROR - Cannot open input file\n");
		exit(-1);
	}
	printf("c %-28s:%s\n","INSTANCE",argv[argc-1]);
	printf("c %-28s:%d\n","SEED",seed);
	//printf("\nINSTANCE: %s\n",argv[argc-1]);

}

void parseParametersComp(int argc,char *argv[]) {
	numTries = DEFAULT_NUMTRIES;
	cutoff = DEFAULT_CUTOFF_FLIPS;
	targetCost = DEFAULT_TARGETCOST;
	printSol = DEFAULT_PRINTSOL;

	fpInput = fopen(argv[1], "r");
	if (fpInput == NULL) {
		fprintf(stderr, "ERROR - Cannot open input file\n");
		exit(-1);
	}

	if (argc > 2) scanInt(argc, argv, 2, &seed);
}

void scanInt(int argc, char *argv[], int i, int *varptr) {
	if (i>=argc || sscanf(argv[i],"%i",varptr)!=1){
		fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
		exit(-1);
	}
}


void scanFloat(int argc, char *argv[], int i, float *varptr) {
	if (i>=argc || sscanf(argv[i],"%f",varptr)!=1){
		fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
		exit(-1);
	}
}

int compareAbsInts(const void *a, const void *b) {
	int *va = (int *) a;
	int *vb = (int *) b;
	return ( abs(*va) - abs(*vb) );
}

void initInstance() {
	int i, j, lit, freestore, lastc, nextc;
	int *storeptr = 0;
	char buf[100];
	int neighbours[MAXATOM+1];
	int numNeighbours;
	int numocc, c, var, k;
	int *occptr, *ptr;

	lastc = '#';
	while (lastc != 'p') {
		while ((lastc = fgetc(fpInput)) == 'c') {
			while ((nextc = fgetc(fpInput)) != EOF && nextc != '\n');
		}
	}

	ungetc(lastc, fpInput);
	if (fscanf(fpInput, "p %s %i %i", buf, &numAtoms, &numClauses) != 3) {
		fprintf(stderr,"ERROR - Bad input file!\n");
		exit(-1);
	}
	if (strcmp("cnf", buf) != 0 ) {
		fprintf(stderr, "ERROR - Bad input file\n");
		exit(-1);
	}
	if (numAtoms > MAXATOM) {
		fprintf(stderr,"ERROR - too many atoms\n");
		exit(-1);
	}

	freestore = 0;
	numLiterals = 0;
	for (i = 0; i < MAXLITERAL+1; i++) numOccurence[i] = 0;

	minClauseSize = maxClauseSize = NOVALUE;
	for (i = 0; i < numClauses; i++) {
		size[i] = -1;
		if (freestore < MAXLENGTH) {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
		}
		resClause[i] = storeptr;
		//		cost[i] = 1;

		do {
			size[i]++;
			if (size[i] > MAXLENGTH) {
				fprintf(stderr, "ERROR - clause too long\n");
				exit(-1);
			}
			if (fscanf(fpInput, "%i ", &lit) != 1) {
				fprintf(stderr, "ERROR - Bad input file\n");
				exit(-1);
			}
			if (lit != 0) {
				*(storeptr++) = lit;
				freestore--;
				numLiterals++;
				numOccurence[lit+MAXATOM]++;
			}
		} while (lit != 0);
		qsort(resClause[i], size[i], sizeof(int*), compareAbsInts);
		if ( (size[i] < minClauseSize) || (minClauseSize == NOVALUE) ) minClauseSize = size[i];
		if ( (size[i] > maxClauseSize) || (maxClauseSize == NOVALUE) ) maxClauseSize = size[i];
	}
	fclose(fpInput);
	if (size[0] == 0) {
		fprintf(stderr,"ERROR - incorrect problem format or extraneous characters\n");
		exit(-1);
	}

	for (i = 0; i < MAXLITERAL+1; i++) {
		if (freestore < numOccurence[i]) {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
		}
		occurence[i] = storeptr;
		freestore -= numOccurence[i];
		storeptr += numOccurence[i];
		numOccurence[i] = 0;
	}

	for (i = 0; i < numClauses; i++) {
		for (j = 0; j < size[i]; j++) {
			occurence[Lit(i,j)+MAXATOM][numOccurence[Lit(i,j)+MAXATOM]] = i;
			numOccurence[Lit(i,j)+MAXATOM]++;
		}
	}

	for (i = 1; i < numAtoms+1; i++) neighbours[i] = 0;

	for (i = 1; i < numAtoms+1; i++) {
		numNeighbours = 0;

		numocc = numOccurence[MAXATOM-i];
		occptr = occurence[MAXATOM-i];
		for (j = 0; j < numocc ; j++) {
			c = *(occptr++);
			for (k = 0, ptr = resClause[c]; k < size[c]; k++, ptr++) {
				var = abs(*ptr);
				if ((neighbours[var] != i) && (var != i)) {
					neighbours[var] = i;
					numNeighbours++;
				}
			}
		}

		numocc = numOccurence[MAXATOM+i];
		occptr = occurence[MAXATOM+i];
		for (j = 0; j < numocc ; j++) {
			c = *(occptr++);
			for (k = 0, ptr = resClause[c]; k < size[c]; k++, ptr++) {
				var = abs(*ptr);
				if ((neighbours[var] != i) && (var != i)) {
					neighbours[var] = i;
					numNeighbours++;
				}
			}
		}

		if (freestore < ++numNeighbours) {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
		}
		neighbourVar[i] = storeptr;
		freestore -= numNeighbours;

		for (j = 1; j < numAtoms+1; j++) {
			if (neighbours[j] == i) {
				*(storeptr++) = j;
				if (--numNeighbours == 1) {
					*(storeptr++) = NOVALUE;
					break;
				}
			}
		}
	}
}

void initInstance_stdin() {
	int i, j, lit, freestore, lastc, nextc;
	int *storeptr = 0;
	char buf[100];
	int neighbours[MAXATOM+1];
	int numNeighbours;
	int numocc, c, var, k;
	int *occptr, *ptr;

	lastc = '#';
	while (lastc != 'p') {
		while ((lastc = getchar()) == 'c') {
			while ((nextc = getchar()) != EOF && nextc != '\n');
		}
	}

	ungetc(lastc, stdin);
	if (scanf("p %s %i %i", buf, &numAtoms, &numClauses) != 3) {
		fprintf(stderr,"ERROR - Bad input file!\n");
		exit(-1);
	}
	if (strcmp("cnf", buf) != 0 ) {
		fprintf(stderr, "ERROR - Bad input file\n");
		exit(-1);
	}
	if (numAtoms > MAXATOM) {
		fprintf(stderr,"ERROR - too many atoms\n");
		exit(-1);
	}

	freestore = 0;
	numLiterals = 0;
	for (i = 0; i < MAXLITERAL+1; i++) numOccurence[i] = 0;

	minClauseSize = maxClauseSize = NOVALUE;
	for (i = 0; i < numClauses; i++) {
		size[i] = -1;
		if (freestore < MAXLENGTH) {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
		}
		resClause[i] = storeptr;
		//		cost[i] = 1;

		do {
			size[i]++;
			if (size[i] > MAXLENGTH) {
				fprintf(stderr, "ERROR - clause too long\n");
				exit(-1);
			}
			if (scanf("%i ", &lit) != 1) {
				fprintf(stderr, "ERROR - Bad input file\n");
				exit(-1);
			}
			if (lit != 0) {
				*(storeptr++) = lit;
				freestore--;
				numLiterals++;
				numOccurence[lit+MAXATOM]++;
			}
		} while (lit != 0);
		qsort(resClause[i], size[i], sizeof(int*), compareAbsInts);
		if ( (size[i] < minClauseSize) || (minClauseSize == NOVALUE) ) minClauseSize = size[i];
		if ( (size[i] > maxClauseSize) || (maxClauseSize == NOVALUE) ) maxClauseSize = size[i];
	}
	if (size[0] == 0) {
		fprintf(stderr,"ERROR - incorrect problem format or extraneous characters\n");
		exit(-1);
	}

	for (i = 0; i < MAXLITERAL+1; i++) {
		if (freestore < numOccurence[i]) {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
		}
		occurence[i] = storeptr;
		freestore -= numOccurence[i];
		storeptr += numOccurence[i];
		numOccurence[i] = 0;
	}

	for (i = 0; i < numClauses; i++) {
		for (j = 0; j < size[i]; j++) {
			occurence[Lit(i,j)+MAXATOM][numOccurence[Lit(i,j)+MAXATOM]] = i;
			numOccurence[Lit(i,j)+MAXATOM]++;
		}
	}

	for (i = 1; i < numAtoms+1; i++) neighbours[i] = 0;

	for (i = 1; i < numAtoms+1; i++) {
		numNeighbours = 0;

		numocc = numOccurence[MAXATOM-i];
		occptr = occurence[MAXATOM-i];
		for (j = 0; j < numocc ; j++) {
			c = *(occptr++);
			for (k = 0, ptr = resClause[c]; k < size[c]; k++, ptr++) {
				var = abs(*ptr);
				if ((neighbours[var] != i) && (var != i)) {
					neighbours[var] = i;
					numNeighbours++;
				}
			}
		}

		numocc = numOccurence[MAXATOM+i];
		occptr = occurence[MAXATOM+i];
		for (j = 0; j < numocc ; j++) {
			c = *(occptr++);
			for (k = 0, ptr = resClause[c]; k < size[c]; k++, ptr++) {
				var = abs(*ptr);
				if ((neighbours[var] != i) && (var != i)) {
					neighbours[var] = i;
					numNeighbours++;
				}
			}
		}

		if (freestore < ++numNeighbours) {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
		}
		neighbourVar[i] = storeptr;
		freestore -= numNeighbours;

		for (j = 1; j < numAtoms+1; j++) {
			if (neighbours[j] == i) {
				*(storeptr++) = j;
				if (--numNeighbours == 1) {
					*(storeptr++) = NOVALUE;
					break;
				}
			}
		}
	}
}

/*
void initInstance() {
	int i, j, lit, freestore, lastc, nextc;
	int *storeptr = 0;
	char buf[100];

	//=== Fix variables for duplicate literal clauses
	//=== (duplicate literals caused the SAPS solver to fail)
	int duplicate, redundant, numRedundant = 0;

	while ((lastc = getchar()) == 'c') {
		while ((nextc = getchar()) != EOF && nextc != '\n');
	}

	ungetc(lastc,stdin);
	if (scanf("p %s %i %i", buf,&numAtoms,&numClauses) != 3) {
		fprintf(stderr,"ERROR - Bad input file!\n");
		exit(-1);
	}
	if (strcmp("cnf", buf) != 0 ) {
		fprintf(stderr, "ERROR - Bad input file\n");
		exit(-1);
	}
	if (numAtoms > MAXATOM) {
		fprintf(stderr,"ERROR - too many atoms\n");
		exit(-1);
	}

	freestore = 0;
	numLiterals = 0;
	for (i = 0; i < MAXLITERAL+1; i++) numOccurence[i] = 0;

	for (i = 0;i < numClauses;i++) {
		size[i] = -1;
		if (freestore < MAXLENGTH) {
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
		}
		resClause[i] = storeptr;
		cost[i] = 1;

		duplicate = redundant = FALSE;
		do {
			size[i]++;
			if (size[i] > MAXLENGTH) {
				fprintf(stderr, "ERROR - clause too long\n");
				exit(-1);
			}
			if (scanf("%i ", &lit) != 1) {
				fprintf(stderr, "ERROR - Bad input file\n");
				exit(-1);
			}
			if (redundant) size[i]--;
			else if(lit != 0) {
				//== Check for duplicate literals
				duplicate = FALSE;
				for(j = 0; j < size[i]; ++j)	{
					if (abs(lit) == abs(resClause[i][j])) {
						if(lit == resClause[i][j]) duplicate = TRUE;
						else redundant = TRUE;
					}
				}
				if (!duplicate && !redundant) {
 *(storeptr++) = lit;
					freestore--;
					numLiterals++;
					numOccurence[lit+MAXATOM]++;
				} else {
					size[i]--;
					if(duplicate) printf("Removing duplicate lit %d from clause %d\n", lit, i);
				}
			}
		} while (lit != 0);
		if (redundant) { //=== Remove redundant clause
			for(j = 0; j < size[i]; ++j) 		 {
				storeptr--;
				freestore++;
				numLiterals--;
				numOccurence[resClause[i][j]+MAXATOM]--;
			}
			printf("Removing redundant clause %d\n", i);
			numRedundant++;
			i--;
			numClauses--;
		}
	}
	if (numRedundant)
		printf("Of %d original clauses %d were redundant\n", numClauses+numRedundant, numRedundant);
	if (size[0] == 0) {
		fprintf(stderr,"ERROR - incorrect problem format or extraneous characters\n");
		exit(-1);
	}

	for (i = 0; i < MAXLITERAL+1; i++) {
		if (freestore < numOccurence[i]){
			storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
			freestore = STOREBLOCK;
		}
		occurence[i] = storeptr;
		freestore -= numOccurence[i];
		storeptr += numOccurence[i];
		numOccurence[i] = 0;
	}

	for (i = 0; i < numClauses; i++) {
		for (j = 0; j < size[i]; j++) {
			occurence[Lit(i,j)+MAXATOM][numOccurence[Lit(i,j)+MAXATOM]] = i;
			numOccurence[Lit(i,j)+MAXATOM]++;
		}
	}
}
 */

void printHeader() {
	printf("c +------------------------------+\n");
	printf("c | gNovelty+, version 1.0       |\n");
	printf("c +------------------------------+\n");
	printf("c gnovelty+ -h for help\nc \n");
}


void printStatsHeader() {

	printf("c Problem instance: numAtoms = %i, numClauses = %i, numLiterals = %i\n",numAtoms, numClauses, numLiterals);
	printf("c General parameters: seed=%d, cutoff=%d, targetCost = %d, print solution: ", seed, cutoff, targetCost);
	printf( printSol?"yes":"no");
	printf("\nc gNovelty+ parameters: noise = %d, walkprob = %d, smoothprob = %d", noise, walkProb, smoothProb);
	printf("\nc tries = %d\nc \n",numTries);

	//printf("  best    flips     final     total     total\n");
	//printf("  cost    until    #unsat     flips   seconds\n");
	//printf("  this     best      this      this      this\n");
	//printf("   try     soln       try      try        try\n");
	fflush(stdout);
}


void printStatsEndTry() {
	printf("c %-28s:%d\n","BARRIER",barrier);
	printf("c %-28s:%d\n","FREEDEGREE",freeDegree);
	printf("c %-28s:%d\n","SEED",seed);
	printf("c %-28s:%d\n","FLIPS",flip);
	if (numFalse==0)
		printf("c RESULT: GNOVELTY+ (M%3d): S :%-10.2lf\n",marchCount,expertime);
	else
		printf("c RESULT: GNOVELTY+ (M%3d): U :%-10.2lf\n",marchCount,expertime);

	//		printf("LOWCOST:     %12d\n",lowCost);
	//		printf("LOWFLIPS:    %12d\n",lowFlips);
	//		printf("TOTALFLIP: %12ld\n",flip-nullFlip);
	//		printf("TRYTIME:      %10.2lf\n",trytime);
	//		printf("%10d %10d %10d %10ld %10.2lf", lowCost, lowFlips, numFalse, flip - nullFlip, trytime);
	//		printf("\n");
	//printFinalStatsComp();
	fflush(stdout);
}

void printStatsEndTryM() {
	printf("c %-28s:%d\n","BARRIER",barrier);
	printf("c %-28s:%d\n","FREEDEGREE",freeDegree);
	printf("c %-28s:%d\n","SEED",seed);
	printf("c %-28s:%d\n","FLIPS",flip);
	printf("c RESULT: MARCH_KS  (M%3d): S :%-10.2lf\n",marchCount,expertime);
	//printf("c %-28s:%-10.2lf\n","RESULT: MARCHKS: S",expertime);
	//			printf("LOWCOST:     %12d\n",lowCost);
	//			printf("LOWFLIPS:    %12d\n",lowFlips);
	//			printf("TOTALFLIP: %12ld\n",flip-nullFlip);
	//			printf("TRYTIME:      %10.2lf\n",trytime);
	//			printf("%10d %10d %10d %10ld %10.2lf", lowCost, lowFlips, numFalse, flip - nullFlip, trytime);
	//			printf("\n");
	//printFinalStatsComp();
	fflush(stdout);
}


void updateStatsEndTry() {
	totalFlips += flip;

	if (numFalse <= targetCost) {
		saveSolution();
		numSuccessTries++;
		totalSuccessFlips += flip;
		totalSuccessNullFlips += nullFlip;
	}
	if (computeCurrentCost() != numFalse) {
		fprintf(stderr, "Program error, verification of assignment cost fails!\n");
		exit(-1);
	}
	//printFinalStatsComp();
}


void updateStatsEndFlip() {
	if (numFalse < lowCost) {
		lowCost = numFalse;
		lowFlips = flip - nullFlip;

		//	fprintf(stderr, "New best %5d at flip# %10d\n", lowCost, flip);
	}
}

void printFinalStats(void) {
	printf("\ntotal elapsed seconds = %f\n", expertime);
	printf("\ntotal TotalNullflips = %ld\n", totalNullFlips);
	printf("average flips per second = %ld\n", (long)((totalFlips - totalNullFlips)/expertime));
	printf("number solutions found = %d\n", numSuccessTries);
	printf("final success rate = %f\n", ((double)numSuccessTries * 100.0)/numTries);
	printf("average length successful tries including nulls = %li\n",
			numSuccessTries ? (totalSuccessFlips/numSuccessTries) : 0);
	if (numSuccessTries > 0) {
		printf("mean flips for successful tries = %.6f\n",
				numSuccessTries ? ((double) (totalSuccessFlips - totalSuccessNullFlips)/numSuccessTries) : 0);
	}

	if (numSuccessTries > 0) {
		printf("ASSIGNMENT FOUND\n");
		//	if (printSol == TRUE) printSolutionG();
	} else printf("ASSIGNMENT NOT FOUND\n");
}

void printFinalStatsComp() {
	if (numSuccessTries > 0) {
		printf("s SATISFIABLE\n");
		if (printSol == TRUE) printSolutionG();
		printf("c Done in %f seconds\n", expertime);
		exit(10);
	} else {
		printf("s UNKNOWN\n");
		exit(0);
	}
}

CWTYPE computeCurrentCost() {
	int i, j, sat;
	CWTYPE result = 0;

	for (i = 0; i < numClauses; i++) {
		sat = FALSE;
		for (j = 0; j < size[i]; j++) {
			if ( (Lit(i,j) > 0) == atom[Var(i,j)] ){
				sat = TRUE;
				break;
			}
		}
		//if (!sat) result += cost[i];
		if (!sat) result++;
	}
	return result;
}

void saveSolution(void) {
	register int i;

	for (i = 1; i < numAtoms+1; i++)
		solution[i] = atom[i];

	//saveUnitVariableValue();
}


void printSolutionG(void) {
	register int i;

	for (i = 1; i < numAtoms+1; i++)
		printf("v %i\n", solution[i] == 1 ? i : -i);
	printf("v 0\n");
	fflush(stdout);
}


void initRingBuf() { //initialisieren von dem ganzen ding
	int i;
	for(i=0; i<RINGBUFSIZE; i++) {
		ringBuf[i]=-1;

	}
	ringBufValid=0;
	ringBufPointer=0;
	ringBufReader=0;
}

void insertRingBuf(int var) { //insert (is ja klar)
	ringBuf[ringBufPointer]=var;

	ringBufPointer++;
	if(ringBufPointer==RINGBUFSIZE) {
		ringBufPointer=0;
	}
	if(ringBufPointer==ringBufReader) {
		ringBufValid=0;
	}
}

void initReadRingBuf() { //einmal aufrufen, bevor angefangen wird zu lesen.
	//jeder aufruf von readRingBuf() liefert den nächsten eintrag zurück
	ringBufReader=ringBufPointer-1;
	if(ringBufReader<0) {
		ringBufReader=RINGBUFSIZE-1;
	}
	ringBufValid=1;
}

int readRingBuf() { //liest den Puffer rückwärts aus
	if(ringBufValid==0) {
		return -1;
	}
	int value=ringBuf[ringBufReader];

	ringBufReader--;
	if(ringBufReader<0) {
		ringBufReader=RINGBUFSIZE-1;
	}
	if(ringBufReader==ringBufPointer) {
		ringBufValid=0;
	}
	return value;
}



int callMarchKS(int* assignment, int freeAtoms, int** resClause, int marchResult[]){
	//create dummy args for MarchKS

	char* args[2];
	char char1[1];
	char char2[1];
	char1[0] = 'a';
	char2[0] = 'b';
	args[0] = char1;
	args[1] = char2;
	//	printf("c Hybrid::callMarchKS():\n");	fflush(stdout);
	//	printf("c callMarchKS():: The assignment, that March KS checks is:\nc callMarchKS():: ");
	int i;
	int j;
	//	for (i = 1; i < numAtoms+1; i++){
	//		printf("%i", assignment[i]);
	//	}
	//	printf("\n");
	//init the clauses array for MarchKS
	//printf("numClauses: %d",numClauses);
	//printf("numAtoms: %d",numAtoms);
	//printf("freeAtoms: %d",freeAtoms);

	int clauses[numClauses + (numAtoms-freeAtoms)][20];//TODO ACHTUNG!!! wir haben keine klauseln, die laenger sind// = (int**) malloc( sizeof( int* ) * (numClauses + (numAtoms-freeAtoms)) );
	//	printf("c callMarchKS():: The number of ORIGINAL clauses is: %i, positions [0,%i].\n", numClauses,numClauses-1);
	//	if ((numAtoms-freeAtoms) > 0){
	//		printf("c callMarchKS():: The number of NEW (unit) clauses is: %i, positions [%i,%i].\n", (numAtoms-freeAtoms), numClauses, numClauses+(numAtoms-freeAtoms)-1);
	//	} else {
	//		printf("c callMarchKS():: The number of NEW (unit) clauses is: %i.\n", (numAtoms-freeAtoms));
	//	}fflush(stdout);
	int unary = 0;

	//fill the clauses array with clauses
	fflush(stdout);
	for(i=0; i<numClauses + numAtoms; i++) {
		if(i<numClauses) {
			for(j=0; j<size[i]; j++) {
				clauses[i][j]=resClause[i][j];
			}
			clauses[i][j++]=0;
		} else {
			if(assignment[i-numClauses+1] !=2) {
				if(assignment[i-numClauses+1] == 0) {
					clauses[numClauses+unary][0]=-1*(i-numClauses+1); //0815 trick
				} else {
					clauses[numClauses + unary][0]=i-numClauses+1;
				}
				clauses[numClauses + unary][1]=0;
				unary++;
			}
		}
	}
	fflush(stdout);
	int** cheatClauses=(int**)malloc((numClauses + (numAtoms-freeAtoms))*sizeof(int*));
	for(i=0; i<numClauses + (numAtoms-freeAtoms); i++) {
		cheatClauses[i]=clauses[i];
	}
	int result=march(2,args,cheatClauses,numAtoms,(numClauses+(numAtoms-freeAtoms)),marchResult);
	free(cheatClauses);
	return result;


}
