/* MARCH Satisfiability Solver

   Copyright (C) 2001-2004 M. Dufour, M. Heule, J. van Zwieten
   [m.dufour@student.tudelft.nl, marijn@heule.nl, zwieten@ch.tudelft.nl]

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include "common.h"
#include "preselect.h"
#include "equivalence.h"

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <assert.h>

#define MAX_ITERATIONS		20
#define RANK_TRIGGER		0
//#define RANK_TRIGGER		1000000

/* variables */
int *freevarsLookup;
int *EqAndTernaryImpSize;

int* CandidatesSet, nrofCandidates;

double preselect_counter;
int dynamic_preselect_setsize;

inline int compute_max_preselected();

void init_preselection()
{
	int i;

	preselect_counter = 0;

	CandidatesSet = (int*) malloc( sizeof( int ) * (nrofvars + 1) );

        /* Pre-selecting Ranking values */
        Rank = (int*) malloc( sizeof( int ) * ( nrofvars + 1 ) );

        EqAndTernaryImpSize = (int *) malloc( sizeof( int ) * ( 2 * nrofvars + 1 ) );
        for( i = 0; i < ( 2 * nrofvars + 1 ); i++ ) EqAndTernaryImpSize[ i ] = 0;
	EqAndTernaryImpSize += nrofvars;

	initial_freevars = freevars;

	Rank_trigger = RANK_TRIGGER;
}

void reinit_freevars()
{
	if( freevarsArray != NULL )  free( freevarsArray );
	if( freevarsLookup != NULL ) free( freevarsLookup );

	init_freevars();
}

void init_freevars()
{
	int i, *Reductions;

	if( kSAT_flag )	Reductions = big_occ;
	else 		Reductions = TernaryImpSize;

    	freevarsArray  = (int*) malloc( sizeof( int ) * (nrofvars + 1) );
	freevarsLookup = (int*) malloc( sizeof( int ) * (nrofvars + 1) );

        freevars = 0;
#ifdef EQ
//        for( i = 0; i < nrofceq; i++ )
//	    if( CeqSizes[ i ] > 2 )
//		VeqDepends[ CeqDepends[ i ] ] = EQUIVALENT;
#endif
        for( i = 1; i <= nrofvars; i++ )
        {
	    freevarsLookup[ i ] = -1;

	    if( (Reductions[ i ] + Reductions[ -i ] == 0) &&
		(BinaryImp[  i ][ 0 ] == bImp_satisfied[  i ]) &&
		(BinaryImp[ -i ][ 0 ] == bImp_satisfied[ -i ]) &&
		(Veq[ i ][ 0 ] == 1 ) )
		    continue;

	    if( (timeAssignments[ i ] < VARMAX) )
            {
		freevarsArray [ freevars  ] = i;
		freevarsLookup[     i     ] = freevars++;
            }
	}

#ifdef DYNAMIC_PRESELECT_SETSIZE
	dynamic_preselect_setsize = check_ternary_clause_density();
#else
	dynamic_preselect_setsize = 0;
#endif
	if( dynamic_preselect_setsize ) {
		if (DOOUTPUT)printf("c dynamic_preselect_setsize :: on\n");
	}
	else{
		if (DOOUTPUT)printf("c dynamic_preselect_setsize :: off\n");
	}
}

int ConstructCandidatesSet( )
{
	int i, _freevar, *_freevarsArray, *Reductions;

	if( kSAT_flag )	Reductions = big_occ;
	else		Reductions = TernaryImpSize;

	nrofCandidates = 0;

        for( _freevarsArray = freevarsArray, i = freevars; i > 0; i-- )
	{
	    _freevar = *(_freevarsArray++);

	    UNFIX(_freevar);
#ifdef EQ
	    if( non_tautological_equivalences )
  	    {
		int j;
		double reduction_sum = 0;
		if( Veq[ _freevar ][ 0 ] > 1 )
	    	    for( j = Veq[ _freevar ][ 0 ] - 1; j > 0; j-- )
			reduction_sum += lengthWeight[ CeqSizes[ Veq[ _freevar ][ j ] ] - 1 ];

		EqAndTernaryImpSize[  _freevar ] = Reductions[  _freevar ] + (int) reduction_sum;
		EqAndTernaryImpSize[ -_freevar ] = Reductions[ -_freevar ] + (int) reduction_sum;
	    }
#endif
	    if( _freevar > original_nrofvars ) continue;

	    if( VeqDepends[ _freevar ] == EQUIVALENT ) continue;
//	    if( VeqDepends[ _freevar ] != INDEPENDENT ) continue;
#ifdef GLOBAL_AUTARKY
	    if( (depth > 0) &&
		(TernaryImpReduction[  _freevar ] == 0) &&
	        (TernaryImpReduction[ -_freevar ] == 0) )
		continue;
#endif
	    CandidatesSet[ nrofCandidates++ ] = _freevar;
	}
	return nrofCandidates;
}

void ComputeRank()
{
	int i, j, _freevar, *bImp, *_freevarsArray, *Reductions;
	int pos, neg;
/*
	for( i = 1; i <= nrofvars; i++ )
	{
	    if( TernaryImpSize[ i] != big_occ[ i] ) printf("TIS != big_occ %i %i at node %i\n",  TernaryImpSize[ i], big_occ[ i], nodeCount );
	    if( TernaryImpSize[-i] != big_occ[-i] ) printf("TIS != big_occ %i %i at node %i\n",  TernaryImpSize[-i], big_occ[-i], nodeCount );
	}
*/
#ifdef EQ
	if( non_tautological_equivalences )
	    Reductions = EqAndTernaryImpSize;
	else
#endif

	if( kSAT_flag ) Reductions = big_occ;
	else		Reductions = TernaryImpSize;

	_freevarsArray = CandidatesSet;
	for( i = nrofCandidates; i > 0; i-- )
	{
	    _freevar = *(_freevarsArray++);
	    pos = Reductions[  _freevar ] + 1;
	    bImp = BinaryImp[ _freevar ] + 2;
	    for(j = bImp[ -2 ] - 1; --j; )
		pos += Reductions[ -*(bImp++) ];
	    if ( pos > 0x00004FFFF ) pos = 0x00004FFFF;

	    neg = Reductions[ -_freevar ] + 1;
	    bImp = BinaryImp[ -_freevar ] + 2;
	    for(j = bImp[ -2 ] - 1; --j; )
		neg += Reductions[ -*(bImp++) ];
	    if ( neg > 0x00004FFFF ) neg = 0x00004FFFF;

	    Rank[ _freevar ] = pos * neg;
	}
}

void RealisePreselectedSet( )
{
        int i, lit, som, *_freevarsArray, max_preselected, iteration_counter;
	double gemiddelde;

#ifndef LOCAL_AUTARKY
	if( dynamic_preselect_setsize ) max_preselected = compute_max_preselected();
	else				max_preselected = (int) (percent * 0.01 * freevars );
	if( max_preselected < 10 ) 	max_preselected = 10;
#endif
        lookaheadArrayLength = 0;
	_freevarsArray = CandidatesSet;
        for( i = nrofCandidates, som = 0; i > 0; i-- )
	{
	    lit = *(_freevarsArray++);

	    if( Rank[ lit ] > Rank_trigger )
	    {
	       	som += Rank[ lit ];
               	lookaheadArray[ lookaheadArrayLength++ ] = lit;
            }
	}

	if( lookaheadArrayLength <= 5 )
	{
            lookaheadArrayLength = 0;
	    _freevarsArray = CandidatesSet;
            for( i = nrofCandidates, som = 0; i > 0; i-- )
	    {
	    	lit = *(_freevarsArray++);

	    	som += Rank[ lit ];
            	lookaheadArray[ lookaheadArrayLength++ ] = lit;
	    }
	}

//	assert( lookaheadArrayLength > 0 );
#ifndef LOCAL_AUTARKY
	iteration_counter = 0;
	while( ((lookaheadArrayLength/2) >= max_preselected) && (iteration_counter < MAX_ITERATIONS) )
	{
            gemiddelde = som / lookaheadArrayLength - 1;

            for( i = 0, som = 0; i < lookaheadArrayLength; )
  	    	if( ( (int) Rank[ lookaheadArray[ i ] ] ) >= gemiddelde )
                {
                    som += (int) Rank[ lookaheadArray[ i++ ] ];
		}
                else    lookaheadArray[ i ] = lookaheadArray[ --lookaheadArrayLength ];

	    assert( lookaheadArrayLength > 0 );

	    iteration_counter++;
	}
#endif
	qsort( lookaheadArray, lookaheadArrayLength, sizeof(int), RankCompare );
#ifndef LOCAL_AUTARKY
	if( lookaheadArrayLength > max_preselected )
	    lookaheadArrayLength = max_preselected;
#endif
}

void ConstructPreselectedSet( )
{
	ComputeRank( );
	RealisePreselectedSet( );
}

int PreselectAll( )
{
	int i, _freevar, *Reductions;

	if( kSAT_flag )	Reductions = big_occ;
	else		Reductions = TernaryImpSize;

        lookaheadArrayLength = 0;

        for( i = 0; i < freevars; i++ )
	{
	    _freevar = freevarsArray[ i ];
	    if( (Reductions[ _freevar ] > 0) || (Reductions[ -_freevar ] > 0) ||
		(BinaryImp[  _freevar ][ 0 ] > bImp_satisfied[  _freevar ]) ||
		(BinaryImp[ -_freevar ][ 0 ] > bImp_satisfied[ -_freevar ]) ||
		(Veq[ _freevar ][ 0 ] > 1 ) )
        	lookaheadArray[ lookaheadArrayLength++ ] = _freevar;
        	CandidatesSet [ nrofCandidates++       ] = _freevar;
	}

	//printf("Preselected 'ALL' %i\n", nrofCandidates);

	ComputeRank();

	qsort( lookaheadArray, lookaheadArrayLength, sizeof( int ), RankCompare );

	return lookaheadArrayLength;
}

int RankCompare(const void *ptrA, const void *ptrB)
{
	return ( Rank[ *(int *)ptrA ] - Rank[ *(int *)ptrB ] ) > 0 ? -1 : 1;
}

int compute_max_preselected( )
{
	if( depth <= MAX_FULL_LOOK_DEPTH )
	{
	    preselect_counter += (double) freevars;
	    return freevars;
	}

	preselect_counter += (double) (5 + 7 * forced_literals );

	return (int) (preselect_counter / nodeCount);
}

void reduce_freevars( int nrval )
{
	int tmp, nr;

	if( freevarsLookup[ NR(nrval) ] == -1 ) return;

	tmp = freevarsArray[ --freevars ];
	nr = NR(nrval);

   	freevarsArray[ freevarsLookup[ nr ] ] = tmp;
   	freevarsArray[ freevars ] = nr;

   	freevarsLookup[ tmp ] = freevarsLookup[ nr ];
   	freevarsLookup[  nr ] = freevars;

	Rank[ nr ] = 0;
}

int check_ternary_clause_density()
{
	int i, j, ternary_sum, freevars_sum;
	int *occurences;

	if( kSAT_flag )	occurences = big_occ;
	else		occurences = TernaryImpSize;

	ternary_sum 	= 0;
	freevars_sum = freevars;

        for( i = 0; i < freevars; i++ )
        {
	    j = freevarsArray[ i ];
            if ( VeqDepends[ j ] == EQUIVALENT )
              	freevars_sum--;
            ternary_sum += occurences[ j ] + occurences[ -j ];
	}

	if( freevars > freevars_sum * 1.5 )
	{
		if (DOOUTPUT)
		printf("c many dependent variables -> full lookahead\n");
		return 1;
	}

	ternary_sum = ternary_sum / 3;

	return ((ternary_sum + nrofceq) > 3*(freevars_sum+nrofvars-original_nrofvars))? 0 : 1;
}

#ifdef EQ
void clearEqAndTernaryImpSize( const int nrval )
{
	EqAndTernaryImpSize[  nrval ] = 0;
	EqAndTernaryImpSize[ -nrval ] = 0;
}
#endif

void disposePreselection()
{
	free( CandidatesSet );
	
	if( EqAndTernaryImpSize != NULL ) { EqAndTernaryImpSize -= nrofvars; free( EqAndTernaryImpSize ); }
	if( Rank                != NULL ) free( Rank );
	if( freevarsArray       != NULL ) free( freevarsArray  );
	if( freevarsLookup      != NULL ) free( freevarsLookup );
}

