/* MARCH Satisfiability Solver

   Copyright (C) 2001-2005 M. Dufour, M. Heule, J. van Zwieten
   [m.dufour@student.tudelft.nl, marijn@heule.nl, zwieten@ch.tudelft.nl]

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

*/

//#define LOCAL_AUTARKY
//#define DETECT_COMPONENTS
#define COMPENSATION_RESOLVENTS

#define SAT_INCREASE	1000

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#include "common.h"
#include "blocked.h"
#include "distribution.h"
#include "rootlook.h"
#include "solver.h"
#include "equivalence.h"
#include "preselect.h"
#include "memory.h"
#include "translator.h"
#include "lookahead.h"
#include "progressBar.h"
#include "parser.h"
#include "localbranch.h"

#define CONTINUE	1
#define FINISHED	0

#define IS_REDUCED_TIMP( __a, __b )	((timeAssignments[__a] < NARY_MAX) && (timeAssignments[__b] == (NARY_MAX+1)) )

#ifdef DISTRIBUTION
  #define LEFT_CHILD	 records[record_index].child[branch_literal > 0]
  #define RIGHT_CHILD	 records[record_index].child[branch_literal < 0]
#endif

int *tmpEqImpSize;

int *TernaryImpTable, *TernaryImpLast;
int current_bImp_stamp, *bImp_stamps;

int *newbistack, *newbistackp, newbistackSize;
int *substack,   *substackp,   substackSize;

int analyze_autarky();

int nrofforced;

int *autarky_level;

unsigned int solution_bin = 0, solution_bits = 31;

#ifdef DISTRIBUTION
int skip_flag = 0;
#endif
#ifdef SUBTREE_SIZE
int path_length;
#endif
#ifdef PLOT
int plotCount = 0;
#endif
#ifdef CUT_OFF
int last_SAT_bin = -1;
#endif

int currentNodeNumber = 1, UNSATflag = 0;

#define STAMP_IMPLICATIONS( _nrval ) \
{ \
    bImp = BIMP_START(_nrval); \
    current_bImp_stamp++; \
    bImp_stamps[ _nrval ] = current_bImp_stamp; \
    for (i = BIMP_ELEMENTS; --i; ) \
	bImp_stamps[ *(bImp++) ] = current_bImp_stamp; \
}

inline void push_stack_blocks( )
{
	PUSH( r, STACK_BLOCK );
	PUSH( imp, STACK_BLOCK );
        PUSH( bieq, STACK_BLOCK );
        PUSH( subsume, STACK_BLOCK );
	current_node_stamp++;
}

#ifdef PROGRESS_BAR
#define NODE_START( ) \
{ \
	push_stack_blocks( ); \
	pb_descend( ); \
}
#define NODE_END( ) \
{ \
	backtrack(); \
	pb_climb(); \
}
#else
#define NODE_START( ) \
{ \
	push_stack_blocks( ); \
}
#define NODE_END( ) \
{ \
	backtrack(); \
}
#endif \

inline int recursive_solve( )
{
	int _result = 0;

	depth++;
	PUSH( r, STACK_BLOCK );
	_result = march_solve_rec();
	depth--;

	return _result;
}

#define UPDATE_BIN( ) \
{\
    if( depth <= solution_bits ) solution_bin ^= (unsigned int) 1 << (solution_bits - depth); \
}

/*
	----------------------------------------------------
	------------[ initializing and freeing ]------------
	----------------------------------------------------
*/
void clearTernaryImpReduction( )
{
	int i;
	for( i = 1; i <= nrofvars; i++ )
	{
	    TernaryImpReduction[  i ] = 0;
	    TernaryImpReduction[ -i ] = 0;
	}
}

inline void fill_ternary_implication_arrays( )
{
	int i, j, lit;

        for( i = 0; i < nrofclauses; i++ )
	    if( Clength[ i ] == 3 )
                for( j = 0; j < 3; j++ )
                {
                   lit = Cv[i][j];
                   TernaryImp[lit][ tmpTernaryImpSize[lit]++ ] = Cv[i][(j+1)%3 ];
                   TernaryImp[lit][ tmpTernaryImpSize[lit]++ ] = Cv[i][(j+2)%3 ];
                }
}

int initSolver()
{
	int i, j, _tmp;

	/* initialise global counters */

	current_node_stamp = 1;
	lookDead 	   = 0;
	mainDead 	   = 0;
#ifdef COUNT_SAT
	count_sat	   = 0;
#endif
	/* allocate recursion stack */
	/* tree is max. nrofvars deep and we thus have max. nrofvars STACK_BLOCKS
		 -> 2 * nrofvars should be enough for everyone :)
	*/
	INIT_ARRAY( r       , 3 * nrofvars + 1   );
	INIT_ARRAY( imp     , INITIAL_ARRAY_SIZE );
	INIT_ARRAY( subsume , INITIAL_ARRAY_SIZE );
	INIT_ARRAY( bieq    , INITIAL_ARRAY_SIZE );
	INIT_ARRAY( newbi   , INITIAL_ARRAY_SIZE );
	INIT_ARRAY( sub     , INITIAL_ARRAY_SIZE );

	MALLOC_OFFSET( bImp_satisfied, int, nrofvars, 2 );
	MALLOC_OFFSET( bImp_start,     int, nrofvars, 2 );
	MALLOC_OFFSET( bImp_stamps,    int, nrofvars, 0 );
	MALLOC_OFFSET( node_stamps, tstamp, nrofvars, 0 );

//	init_lookahead();

//	return 1;

        init_lookahead();
        init_preselection();

#ifdef DISTRIBUTION
	init_direction();
#endif

	tmpTernaryImpSize = (int* ) malloc( sizeof(int ) * ( 2*nrofvars+1 ) );
#ifdef TERNARYLOOK
        TernaryImp 	  = (int**) malloc( sizeof(int*) * ( 2*nrofvars+1 ) );
        TernaryImpSize 	  = (int* ) malloc( sizeof(int ) * ( 2*nrofvars+1 ) );

        for( i = 0; i <= 2 * nrofvars; i++ )
	{
	    tmpTernaryImpSize[ i ] = 0;
	    TernaryImpSize   [ i ] = 0;
	}

	for( i = 0; i < nrofclauses; i++ )
	    if( Clength[ i ] == 3 )
		for( j = 0; j < 3; j++ )
		    TernaryImpSize[ Cv[ i ][ j ] + nrofvars ]++;

        for( i = 0; i <= 2 * nrofvars; i++ )
	    TernaryImp[ i ] = (int*) malloc(sizeof(int)*(4*TernaryImpSize[i]+4));

        tmpTernaryImpSize 	+= nrofvars;
        TernaryImp 		+= nrofvars;
	TernaryImpSize 		+= nrofvars;

	fill_ternary_implication_arrays();

        for( i = -nrofvars; i <= nrofvars; i++ )
	    tmpTernaryImpSize[ i ] = 4 * TernaryImpSize[ i ] + 2;

//	branch_on_dummies_from_long_clauses();

	while( AddTernaryResolvents() );

        for( i = -nrofvars; i <= nrofvars; i++ )
	    free( TernaryImp[ i ] );

	FREE_OFFSET( TernaryImp     );
	FREE_OFFSET( TernaryImpSize );
#else
        tmpTernaryImpSize 	+= nrofvars;
#endif
#ifdef ROOTLOOK
	if( root_iterative_lookahead() == UNSAT ) return UNSAT;
#endif
	/* initialise global datastructures */

        TernaryImp 	= (int**) malloc( sizeof(int*) * ( 2*nrofvars+1 ) );
        TernaryImpSize 	= (int* ) malloc( sizeof(int ) * ( 2*nrofvars+1 ) );
        TernaryImpLast 	= (int* ) malloc( sizeof(int ) * ( 2*nrofvars+1 ) );
        TernaryImpTable	= (int* ) malloc( sizeof(int ) * ( 6*nrofclauses+1 ) );

	TernaryImp	       += nrofvars;
	TernaryImpSize	       += nrofvars;
	TernaryImpLast	       += nrofvars;

	simplify_formula(); //kan die ook naar onder?

	for( i = -nrofvars; i <= nrofvars; i++ )
	{
	    tmpTernaryImpSize[ i ] = 0;
	    TernaryImpSize   [ i ] = 0;
	    bImp_satisfied   [ i ] = 2;	//waarom staat dit hier?
	}

	for( i = 0; i < nrofclauses; i++ )
	    if( Clength[ i ] == 3 )
		for( j = 0; j < 3; j++ )
		    TernaryImpSize[ Cv[ i ][ j ] ]++;

	_tmp = 0;
        for( i = -nrofvars; i <= nrofvars; i++ )
        {
	    TernaryImp[ i ]     = TernaryImpTable + 2 * _tmp;
            _tmp               += TernaryImpSize[ i ];
	    TernaryImpLast[ i ] = TernaryImpSize[ i ];
        }

	fill_ternary_implication_arrays();

	rebuild_BinaryImp();

	init_freevars();
#ifdef GLOBAL_AUTARKY
	MALLOC_OFFSET( TernaryImpReduction, int, nrofvars, 0 );
#endif
	for( i = 0; i < nrofceq ; i++ )
	    assert( CeqSizes[ i ] != 1 );
#ifdef EQ
	for( i = 0; i < nrofceq ; i++ )
	    if( CeqSizes[ i ] == 2 )
            	DPLL_propagate_binary_equivalence( i );
#endif

	autarky_level = (int*) malloc( sizeof(int) * (nrofvars+1) );
	tmpEqImpSize  = (int*) malloc( sizeof(int) * (nrofvars+1) );
#ifdef DETECT_COMPONENTS
	init_localbranching();
#endif

#ifdef CUT_OFF
	solution_bits = CUT_OFF - 1;
#endif
//	seive_blocked_binaries();

	push_stack_blocks();

#ifdef ROOTLOOK
	return rootlook();
#endif
	return 1;
}

void disposeSolver()
{
	int i;

        destroyLookahead();
#ifdef GLOBAL_AUTARKY
	FREE_OFFSET( TernaryImpReduction );
#endif

#ifdef DISTRIBUTION
	dispose_direction();
#endif
//        free_BinaryImp();

//        for( i = -nrofvars; i <= nrofvars; i++ )
//	    free( TernaryImp[ i ] );

        FREE_OFFSET( TernaryImp        );

        FREE_OFFSET( TernaryImpSize    );
        FREE_OFFSET( TernaryImpLast    );
        FREE_OFFSET( tmpTernaryImpSize );
        FREE_OFFSET( bImp_stamps       );
        FREE_OFFSET( bImp_start       );
        FREE_OFFSET( bImp_satisfied    );
        FREE_OFFSET( node_stamps       );

	free( autarky_level );
	free( tmpEqImpSize  );
	free( TernaryImpTable  );

        free( impstack );
        free( rstack );
        free( subsumestack );
        free( bieqstack );
        free( newbistack );
        free( substack );
}

int propagate_forced_literals()
{
	int _freevars = freevars;

	if( IFIUP( 0, FIX_FORCED_LITERALS ) == UNSAT )
	    return UNSAT;

	nrofforced = _freevars - freevars;
	percentage_forced = 100.0 * nrofforced / (double) _freevars;

//	printf("c lookahead forced %.1f \% variables at depth %i\n", percentage_forced, depth );

	return SAT;
}

#ifdef DISTRIBUTION // naar distribution.c
int distribution_branching()
{
	int _result;

	target_rights  = 0;
	current_rights = 0;
	current_record = 0;

	do
	{
	    assert( target_rights <= jump_depth );

#ifdef SUBTREE_SIZE
	    path_length = 1;
#endif
	    _result = march_solve_rec();
	    if( _result != UNSAT )
		return _result;
#ifdef TIMEOUT
	    if( (int) clock() > CLOCKS_PER_SEC * TIMEOUT )
		return UNKNOWN;
#endif
#ifndef CUT_OFF
	    pb_reset();
#endif
	    target_rights++;

	    current_record = 1;

//	    print_records();
	}
	while( records[1].UNSAT_flag == 0 );

	return UNSAT;

}
#endif

#ifdef DISTRIBUTION
int skip_node( )
{
/*
	if( skip_flag )
	{
//	    printf("%i ", nodeCount );
	    if( current_rights <= target_rights )
	    {
		 skip_flag = 0;
#ifdef SUBTREE_SIZE
		 path_length = depth + 1;
#endif
	    }
	}

	return skip_flag;
*/
	if( depth < jump_depth )
	    if( (target_rights < current_rights) ||
		(target_rights >= (current_rights + jump_depth - depth)) )
		return 1;
	return 0;
}
#endif

int march_solve_rec()
{
	int branch_literal = 0, _result, _percentage_forced;
	int skip_left = 0, skip_right = 0, top_flag = 0;
#ifdef DISTRIBUTION
	int record_index = current_record;

	top_flag = TOP_OF_TREE;

	if( fix_recorded_literals(record_index) == UNSAT )
	    return UNSAT;

	if( record_index == 0 ) record_index = init_new_record( );
#endif
#ifdef TIMEOUT
	if( (int) clock() > CLOCKS_PER_SEC * TIMEOUT )
	    return UNKNOWN;
#endif
#ifdef SUBTREE_SIZE
	path_length++;
#endif
#ifdef CUT_OFF
	if( depth <= CUT_OFF ) last_SAT_bin = -1;

        if( solution_bin == last_SAT_bin )
	{
#ifdef DISTRIBUTION
	    records[ record_index ].UNSAT_flag = 1;
#endif
            return UNSAT;
	}
#endif
#ifdef DETECT_COMPONENTS
	determine_components();
#endif
#ifdef DISTRIBUTION
	branch_literal = records[record_index].branch_literal;

	if( branch_literal == 0 )
#endif
	do
	{
#ifdef LOCAL_AUTARKY
	    int _depth;

	    _depth = analyze_autarky();
	    if( _depth == 0 && DOOUTPUT)
	        printf("c global autarky found at depth %i\n", depth );
	    else if( _depth != depth && DOOUTPUT)
	        printf("c autarky found at depth %i (from %i)\n", depth, _depth );
#endif
	    nodeCount++;

	    if( ConstructCandidatesSet() == 0 )
	    {
	        if( depth > 0 )
	        {
		    if( checkSolution() == SAT ) return verifySolution();
	        }
	    	if( PreselectAll() == 0 )
		    return verifySolution();
	    }
	    else	ConstructPreselectedSet();

	    if( lookahead() == UNSAT )
	    {
	    	lookDead++;
	    	return UNSAT;
	    }

	    if( propagate_forced_literals() == UNSAT )
		return UNSAT;

	    branch_literal = get_signedBranchVariable();

	}
	while( (percentage_forced > 50.0) || (branch_literal == 0) );
#ifdef PLOT
	if( plotCount++ >= 10000 )
	    return UNSAT;

	if (DOOUTPUT) printf("\t%i\t%.3f\t%i\n", plotCount, sum_plot / count_plot, depth );
#endif
	_percentage_forced = percentage_forced;

	NODE_START();

	autarky_level[ depth + 1 ] = FALSE;
#ifdef DISTRIBUTION
	if( top_flag )
	{
	    branch_literal *= -1;
	    current_rights++;
	    UPDATE_BIN();
	}
	skip_left = skip_node();
#endif
	if( (skip_left==0) && IFIUP( branch_literal, FIX_BRANCH_VARIABLE ) )
	{
#ifdef DISTRIBUTION
	    current_record = LEFT_CHILD;
#endif
	    _result = recursive_solve();
#ifdef DISTRIBUTION
	    LEFT_CHILD = current_record;
#endif
	    if( _result == SAT || _result == UNKNOWN ) return _result; }
	else {
#ifdef DISTRIBUTION
//		if( LEFT_CHILD != 0 )  	LEFT_CHILD = 0;
		if( (LEFT_CHILD != 0)  && records[ LEFT_CHILD ].UNSAT_flag == 0 )
		{
		    records[ LEFT_CHILD ].UNSAT_flag = 1;
		    if (DOOUTPUT) printf("c left child %i UNSAT by parent!\n", LEFT_CHILD );
		}
#endif
		PUSH( r, STACK_BLOCK );}

	NODE_END();
#ifdef DISTRIBUTION
	if( top_flag )
	{
	    current_rights--;
	    UPDATE_BIN();
	}
#endif
	if( autarky_level[ depth + 1 ] == TRUE )
	{
		if (DOOUTPUT) printf("c autarky backtrack at depth %i\n", depth);
	    return UNSAT;
	}

	percentage_forced = _percentage_forced;

	NODE_START();

        if( top_flag == 0 )
	{
#ifdef DISTRIBUTION
	    current_rights++;
#endif
	    UPDATE_BIN();
	}
#ifdef DISTRIBUTION
	skip_right = skip_node();
#endif
	if( (skip_right == 0) && IFIUP( -branch_literal, FIX_BRANCH_VARIABLE ) )
	{
#ifdef DISTRIBUTION
	    current_record = RIGHT_CHILD;
#endif
	    _result = recursive_solve();
#ifdef DISTRIBUTION
	    RIGHT_CHILD = current_record;
#endif
	    if( _result == SAT || _result == UNKNOWN ) return _result;}
	else {
#ifdef DISTRIBUTION
		if( (RIGHT_CHILD != 0) && records[ RIGHT_CHILD ].UNSAT_flag == 0 )
		{
		    records[ RIGHT_CHILD ].UNSAT_flag = 1;
		    if (DOOUTPUT) printf("c right child %i UNSAT by parent!\n", RIGHT_CHILD );
		}
#endif
		PUSH( r, STACK_BLOCK );}
	NODE_END();

	if( top_flag == 0 )
	{
#ifdef DISTRIBUTION
	    current_rights--;
#endif
	    UPDATE_BIN();
	}

//	printf("ended node %i at depth %i\n", nodeCount, depth );

#ifdef SUBTREE_SIZE
	if( jump_depth == 0 && current_rights == 0 && path_length - depth > SUBTREE_SIZE ) // testen
//	if( jump_depth == 0 && current_rights == 0 && nodeCount - depth > SUBTREE_SIZE )
	{
	    jump_depth = depth;
	    skip_flag = 1;
	}
#endif
#ifdef DISTRIBUTION
	record_node( record_index, branch_literal, skip_left, skip_right );

	current_record = record_index;
#endif
	return UNSAT;
}

int IFIUP( const int nrval, const int forced_or_branch_flag )
{
	int i, *_forced_literal_array, _forced_literals, *local_fixstackp;

	local_fixstackp = rstackp;
	end_fixstackp   = rstackp;

	currentTimeStamp = BARY_MAX;

	current_bImp_stamp = 1;

	for( i = nrofvars; i >= 1;  i-- )
	{
	    bImp_stamps[  i ] = 0;
	    bImp_stamps[ -i ] = 0;
	}

	if( forced_or_branch_flag == FIX_FORCED_LITERALS )
	{
	   get_forced_literals( &_forced_literal_array, &_forced_literals );
	   for( i = 0; i < _forced_literals; i++ )
	      	if( look_fix_binary_implications(*(_forced_literal_array++)) == UNSAT )
		    { MainDead( local_fixstackp ); return UNSAT; }
	}
#ifdef DISTRIBUTION
	else if( forced_or_branch_flag == FIX_RECORDED_LITERALS )
	{
	   get_recorded_literals( &_forced_literal_array, &_forced_literals );
	   for( i = 0; i < _forced_literals; i++ )
	      	if( look_fix_binary_implications(*(_forced_literal_array++)) == UNSAT )
		    { MainDead( local_fixstackp ); return UNSAT; }
	}
#endif
	else
	{
	 	if( look_fix_binary_implications( nrval ) == UNSAT )
    		    { MainDead( local_fixstackp ); return UNSAT; }
	}

	while( local_fixstackp < end_fixstackp )
		if( DPLL_update_datastructures(*(local_fixstackp++)) == UNSAT )
		    { MainDead( local_fixstackp ); return UNSAT; }

	rstackp = end_fixstackp;

	return SAT;
}

inline void reduce_big_occurences( const int clause_index, const int nrval )
{
#ifdef NO_TRANSLATOR
	int *clauseSet, index;
	int *literals = clause_list[ clause_index ], lit;
	while( *literals != LAST_LITERAL )
	{
	    lit =  *(literals++);
	    if( lit != nrval )
	    {
	        clauseSet = clause_set[ lit ];
	        index = 0; //wellicht niet nodig
	        while( 1 )
	    	{
		    if( *(clauseSet++) == clause_index )
		    {
		        clauseSet[ -1 ] = clause_set[ lit ][ big_occ[ lit ] - 1 ];
		        clause_set[ lit ][ big_occ[ lit ] - 1 ] = LAST_CLAUSE;
		        break;
		    }
		    index++;
		}
	    }
	    big_occ[ lit ]--;
	}
#endif
}

inline int DPLL_update_datastructures( const int nrval )
{
	int i, *bImp;
#ifdef GLOBAL_AUTARKY
	int lit1, lit2;
#endif

#ifdef EQ
	int nr, ceqidx;
	clearEqAndTernaryImpSize( nrval );
	nr = NR( nrval );
        PUSH( sub, STACK_BLOCK );
#endif
//	printf("FIXING %i\n", nrval );

#ifdef TIMEOUT
	if( (int) clock() > CLOCKS_PER_SEC * TIMEOUT )
	    return SAT;
#endif
	unitResolveCount++;
	reduce_freevars( nrval );

        bImp = BIMP_START(-nrval);
        for( i = BIMP_ELEMENTS; --i; )
            bImp_satisfied[ -(*(bImp++)) ]++;

#ifdef GLOBAL_AUTARKY
        tImp = TernaryImp[ nrval ] + 2 * TernaryImpSize[ nrval ];
        for( i = TernaryImpLast[ nrval ] - TernaryImpSize[ nrval ]; i--; )
        {
	    lit1 = *(tImp++);
	    lit2 = *(tImp++);

	    if( IS_REDUCED_TIMP( lit1, lit2 ) )
		TernaryImpReduction[ lit1 ]--;
	    else if( IS_REDUCED_TIMP( lit2, lit1 ) )
                TernaryImpReduction[ lit2 ]--;
        }
#endif
	if( kSAT_flag == 0 )
	{
	    remove_satisfied_implications(  nrval );
	    remove_satisfied_implications( -nrval );
	}
#ifdef GLOBAL_AUTARKY
        tImp = TernaryImp[ -nrval ];
        for( i = tmpTernaryImpSize[ -nrval ]; i--; )
	{
	    TernaryImpReduction[ *(tImp++) ]++;
	    TernaryImpReduction[ *(tImp++) ]++;
	}
#ifdef EQ
	tmpEqImpSize[ nr ] = Veq[ nr ][ 0 ];
	for( i = 1; i < Veq[ nr ][0]; i++ )
        {
	    int j;
            ceqidx = Veq[ nr ][i];
            for( j = 0; j < CeqSizes[ ceqidx ]; j++ )
                TernaryImpReduction[ Ceq[ceqidx][j] ]++;
        }
#endif
#endif
	FIX( nrval, NARY_MAX );

	if( kSAT_flag )
	{
	  int UNSAT_flag, *clauseSet, clause_index;
	  int first_lit, *literals, lit;

	  // REMOVE SATISFIED CLAUSES
	  clauseSet = clause_set[ nrval ];
	  while( *clauseSet != LAST_CLAUSE )
	  {
	    clause_index = *(clauseSet++);
	    if( clause_length[ clause_index ] < SAT_INCREASE - 2 )
		reduce_big_occurences( clause_index, nrval );
	    clause_length[ clause_index ] += SAT_INCREASE;
	  }

	  // REMOVE UNSATISFIED LITERALS
	  UNSAT_flag = 0;
	  clauseSet = clause_set[ -nrval ];
	  while( *clauseSet != LAST_CLAUSE )
	  {
	    clause_index = *(clauseSet++);
	    clause_length[ clause_index ]--;

            if(  clause_length[ clause_index ] == 2 )
            {
		reduce_big_occurences( clause_index, -nrval );
		clause_length[ clause_index ] = SAT_INCREASE;

	        if( UNSAT_flag == 0 )
	        {
                    first_lit = 0;
                    literals = clause_list[ clause_index ];
                    while( *literals != LAST_LITERAL )
                    {
                        lit = *(literals)++;
                        if( IS_NOT_FIXED( lit ) )
                        {
                            if( first_lit == 0 ) first_lit = lit;
                            else
			    {
				UNSAT_flag = !DPLL_add_binary_implications( first_lit, lit );
				goto next_clause;
			    }
                        }
                        else if( !FIXED_ON_COMPLEMENT(lit) ) goto next_clause;
                    }

                    if( first_lit != 0 )  UNSAT_flag = !look_fix_binary_implications( first_lit );
                    else                  UNSAT_flag = 1;
                }
                next_clause:;
	    }
	  }

	  if( UNSAT_flag ) return UNSAT;
	}

	if( kSAT_flag == 0 )
	{
	    int *tImp = TernaryImp[ -nrval ];
            for( i = tmpTernaryImpSize[ -nrval ] - 1; i >= 0; i-- )
                if( DPLL_add_binary_implications( *(tImp++), *(tImp++) ) == UNSAT )
            	    return UNSAT;
	}

#ifdef EQ
        while( Veq[ nr ][ 0 ] > 1 )
        {
            ceqidx = Veq[ nr ][ 1 ];

            fixEq( nr, 1, SGN(nrval));
            PUSH( sub, nrval );

            if( CeqSizes[ ceqidx ] == 2 )
	    {
            	if ( DPLL_propagate_binary_equivalence( ceqidx ) == UNSAT )
               	    return UNSAT;
	    }
	    else if( CeqSizes[ ceqidx ] == 1 )
            {
            	if( look_fix_binary_implications(Ceq[ceqidx][0]*CeqValues[ceqidx]) == UNSAT )
                    return UNSAT;
            }
        }

        while( newbistackp != newbistack )
        {
            POP( newbi, ceqidx );
            if( CeqSizes[ ceqidx ] == 2 )
            	if ( DPLL_propagate_binary_equivalence( ceqidx ) == UNSAT )
                    return UNSAT;
        }
#endif
	return SAT;
}

inline void swap_ternary_implications( const int nrval, const int lit1, const int lit2 )
{
	int i, last, *tImp = TernaryImp[ nrval ];

	last = --TernaryImpSize[ nrval ];
	for( i = last - 1; i >= 0; i-- )
	    if( (tImp[ 2*i ] == lit1) && (tImp[ 2*i + 1 ] == lit2) )
	    {
	    	tImp[ 2*i     ] = tImp[ 2*last     ]; tImp[ 2*last     ] = lit1;
		tImp[ 2*i + 1 ] = tImp[ 2*last + 1 ]; tImp[ 2*last + 1 ] = lit2;
		return;
	    }
}

void remove_satisfied_implications( const int nrval )
{
        int i, lit1, lit2, *tImp = TernaryImp[ nrval ];

        for( i = TernaryImpSize[ nrval ] - 1; i >= 0; i-- )
        {
            lit1 = *(tImp++);
            lit2 = *(tImp++);

	    swap_ternary_implications( lit1, lit2, nrval );
	    swap_ternary_implications( lit2, nrval, lit1 );
        }

        tmpTernaryImpSize[ nrval ] = TernaryImpSize[ nrval ];
        TernaryImpSize   [ nrval ] = 0;
}

int DPLL_propagate_binary_equivalence( const int bieq )
{
        int i, j, ceqsubst;
        int lit1, lit2;
        int value;

        lit1 = Ceq[ bieq ][ 0 ];
        lit2 = Ceq[ bieq ][ 1 ];
        value = CeqValues[ bieq ];

        for( i = 1; i < Veq[ lit1 ][ 0 ]; i++ )
        {
            ceqsubst = Veq[ lit1 ][ i ];
            for( j = 1; j < Veq[ lit2 ][ 0 ]; j++ )
            {
            	if( (ceqsubst == Veq[ lit2 ][ j ]) )
                {
                    fixEq( lit1, i, 1);
                    PUSH( sub, lit1 );

                    fixEq( lit2, j, value);
                    PUSH( sub, lit2 * value );

                    if( CeqSizes[ ceqsubst ] == 0 )
                       	if (CeqValues[ ceqsubst ] == -1 )
                            return UNSAT;

                    if( CeqSizes[ ceqsubst ] == 1 )
                      	if( !look_fix_binary_implications(Ceq[ceqsubst][0] * CeqValues[ceqsubst]) )
                    	    return UNSAT;

                    if( CeqSizes[ ceqsubst ] == 2 )
                     	PUSH( newbi, ceqsubst );

		    i--;
                    break;
                }
            }
        }

        if( (DPLL_add_binary_implications( lit1, -lit2 * value ) &&
	     DPLL_add_binary_implications( -lit1, lit2 * value )) == UNSAT )
                return UNSAT;

        return SAT;
}

inline int DPLL_add_compensation_resolvents( const int lit1, const int lit2 )
{
	int i, lit, *bImp = BIMP_START(lit2);

    	CHECK_NODE_STAMP( -lit1 );
	CHECK_BIMP_UPPERBOUND( -lit1, BinaryImp[ lit2 ][ 0 ] );

	for (i = BIMP_ELEMENTS; --i; )
	{
	    lit = *(bImp++);
	    if( IS_FIXED(lit) ) continue;

	    if( bImp_stamps[ -lit ] == current_bImp_stamp )
		return look_fix_binary_implications( lit1 );
#ifdef COMPENSATION_RESOLVENTS
	    if( bImp_stamps[ lit ] != current_bImp_stamp )
	    {
		CHECK_NODE_STAMP( -lit );
	      	CHECK_BIMP_BOUND  ( -lit );
		ADD_BINARY_IMPLICATIONS( lit, lit1 );
	    }
#endif
    	}
	return UNKNOWN;
}

int DPLL_add_binary_implications( int lit1, int lit2 )
{
	int i, *bImp;

	if( IS_FIXED(lit1) )
	{
	    if( !FIXED_ON_COMPLEMENT(lit1) )	return SAT;
	    else if( IS_FIXED(lit2) )
		    return( !FIXED_ON_COMPLEMENT(lit2) );
	    else    return look_fix_binary_implications(lit2);
	}
	else if( IS_FIXED(lit2) )
	{
	    if( !FIXED_ON_COMPLEMENT(lit2) )	return SAT;
	    else    return look_fix_binary_implications(lit1);
	}

#ifdef BIEQ
	while( (VeqDepends[ NR(lit1) ] != INDEPENDENT) &&
	    (VeqDepends[ NR(lit1) ] != EQUIVALENT) )
		lit1 = VeqDepends[ NR(lit1) ] * SGN(lit1);

	while( (VeqDepends[ NR(lit2) ] != INDEPENDENT) &&
	    (VeqDepends[ NR(lit2) ] != EQUIVALENT) )
		lit2 = VeqDepends[ NR(lit2) ] * SGN(lit2);

	if( lit1 == -lit2 ) return SAT;
	if( lit1 ==  lit2 ) return look_fix_binary_implications(lit1);
#endif

	STAMP_IMPLICATIONS( -lit1 );
	if( bImp_stamps[ -lit2 ] == current_bImp_stamp )
	    return look_fix_binary_implications( lit1 );
	if( bImp_stamps[lit2] != current_bImp_stamp )
	{
	    int _result;

	    bImp_stamps[ BinaryImp[-lit1][ BinaryImp[-lit1][0] - 1] ] = current_bImp_stamp;

	    _result = DPLL_add_compensation_resolvents( lit1, lit2 );
	    if( _result != UNKNOWN )
		return _result;

	    STAMP_IMPLICATIONS( -lit2 );
	    if( bImp_stamps[ -lit1 ] == current_bImp_stamp )
	    	return look_fix_binary_implications( lit2 );

	    _result = DPLL_add_compensation_resolvents( lit2, lit1 );
	    if( _result != UNKNOWN )
		return _result;

	    ADD_BINARY_IMPLICATIONS( lit1, lit2 );
	}
	return SAT;
}

int autarky_stamp( const int nrval )
{
	int i, *tImp, lit1, lit2, flag = 0;

	tImp = TernaryImp[ -nrval ];
	for( i = tmpTernaryImpSize[ -nrval ]; i--; )
	{
	    lit1 = *(tImp++);
	    lit2 = *(tImp++);

	    if( IS_NOT_FIXED(lit1) && IS_NOT_FIXED(lit2) )
	    {
		flag = 1;
		if( VeqDepends[ NR(lit1) ] == DUMMY )
		    autarky_stamp( lit1 );
		else
		    TernaryImpReduction[ lit1 ]++;

		if( VeqDepends[ NR(lit2) ] == DUMMY )
		    autarky_stamp( lit2 );
		else
		TernaryImpReduction[ lit2 ]++;
	    }
	}
	return flag;
}

int analyze_autarky( )
{
#ifdef LOOKAHEAD_ON_DUMMIES
	int *tImp, lit1, lit2;
#endif
	int i, j, nrval, twice;
	int new_bImp_flag, _depth, *_rstackp;

	if( depth == 0 ) return 0;

        for( i = 1; i <= nrofvars; i++ )
        {
            TernaryImpReduction[  i ] = 0;
            TernaryImpReduction[ -i ] = 0;
        }

	new_bImp_flag = 0;
	_rstackp      = rstackp;
	for( _depth = depth; _depth > 0; _depth-- )
	{
//	  if( autarky_level[ _depth ] == TRUE )
//	  {
//		while( *(_rstackp--) == STACK_BLOCK );
//	 	continue;
//	  }

	  for( twice = 1 ; twice <= 2; twice++ )
	    while( *(--_rstackp) != STACK_BLOCK )
	    {
		nrval = *(_rstackp);
#ifndef LOOKAHEAD_ON_DUMMIES
		if( autarky_stamp( nrval ) == 1 )
		    new_bImp_flag = 1;
#else
		tImp = TernaryImp[ -nrval ];
		for( i = tmpTernaryImpSize[ -nrval ]; i--; )
		{
		    lit1 = *(tImp++);
		    lit2 = *(tImp++);

		    if( IS_NOT_FIXED(lit1) && IS_NOT_FIXED(lit2) )
		    {
			new_bImp_flag = 1;
			TernaryImpReduction[ lit1 ]++;
			TernaryImpReduction[ lit2 ]++;
		    }
		}
#endif
	        for( j = 1; j < tmpEqImpSize[NR(nrval)]; j++ )
	        {
	            int ceqsubst = Veq[NR(nrval)][j];
	    	    for( i = 0; Ceq[ceqsubst][i] != NR(nrval); i++ )
		    {
			new_bImp_flag = 1;
			//printf("%i (%i)", Ceq[ ceqsubst][i], _depth );
	   	        TernaryImpReduction[ Ceq[ceqsubst][i] ]++;
		    }
	        }
	    }
	  if( new_bImp_flag == 1 ) break;
	  autarky_level[ _depth ] = TRUE;
	}
	return _depth;
}

void backtrack()
{
	int nrval, varnr, size;

	while( !( *( rstackp - 1 ) == STACK_BLOCK ) )
	{   POP_BACKTRACK_RECURSION_STACK }
	POP_RECURSION_STACK_TO_DEV_NULL

	while( !( *( subsumestackp - 1 ) == STACK_BLOCK ) )
	{
	    POP( subsume, nrval );

	    TernaryImpSize[ TernaryImp[nrval][ 2*TernaryImpSize[nrval]   ] ]++;
	    TernaryImpSize[ TernaryImp[nrval][ 2*TernaryImpSize[nrval]+1 ] ]++;
	    TernaryImpSize[ nrval ]++;
	}
	subsumestackp--;

	while( !( *( rstackp - 1 ) == STACK_BLOCK ) )
	{   POP_BACKTRACK_RECURSION_STACK }
	POP_RECURSION_STACK_TO_DEV_NULL

        while( !( *( bieqstackp - 1 ) == STACK_BLOCK ) )
        {
	    POP( bieq, varnr );
            VeqDepends[ varnr ] = INDEPENDENT;
        }
        bieqstackp--;

	while( !( *( impstackp - 1 ) == STACK_BLOCK ) )
	{
	    POP( imp, size );
	    POP( imp, nrval );
	    BinaryImp[ nrval ][ 0 ] = size;
	}
	impstackp--;
}

void MainDead( int *local_fixstackp )
{
	int nrval;
	mainDead++;

        while( end_fixstackp > local_fixstackp )
        {
	    nrval = *(--end_fixstackp);
	    UNFIX( nrval );
	}
	rstackp = end_fixstackp;
}

inline void restore_big_occurences( const int clause_index, const int nrval )
{
#ifdef NO_TRANSLATOR
	int *literals = clause_list[ clause_index ], lit;
	while( *literals != LAST_LITERAL )
	{
	    lit = *(literals++);
	    if( lit != nrval )
	        clause_set[ lit ][ big_occ[ lit ] ] = clause_index;
	    big_occ[ lit ]++;
	}
#endif
}

void restore_implication_arrays( const int nrval )
{
	int i, *bImp;
#ifdef GLOBAL_AUTARKY
	int lit1, lit2;
#endif
#ifdef EQ
	int ceqsubst, var;
	while( !( *( substackp - 1 ) == STACK_BLOCK ) )
	{
	    POP( sub, var );
	    ceqsubst = Veq[ NR(var) ][ Veq[ NR(var) ][ 0 ]++ ];
	    CeqValues[ ceqsubst ] *= SGN(var);
	    CeqSizes[ ceqsubst ]++;
        }
        substackp--;
#ifdef GLOBAL_AUTARKY
	for( i = 1; i < Veq[NR(nrval)][0]; i++ )
	{
	    int j;
	    ceqsubst = Veq[NR(nrval)][i];
	    for( j = 0; j < CeqSizes[ceqsubst]; j++ )
	   	TernaryImpReduction[ Ceq[ceqsubst][j] ]--;
	}
#endif
#endif
//	printf("UNFIXING %i\n", nrval );

	if( kSAT_flag )
	{
	  int clause_index, *clauseSet;
	  clauseSet = clause_set[ nrval ];
	  while( *clauseSet != LAST_CLAUSE )
	  {
	    clause_index = *(clauseSet++);
	    clause_length[ clause_index ] -= SAT_INCREASE;

	    if( clause_length[ clause_index ] < SAT_INCREASE - 2 )
		restore_big_occurences( clause_index, nrval );
	  }

          clauseSet = clause_set[ -nrval ];
	  while( *clauseSet != LAST_CLAUSE )
          {
            clause_index = *(clauseSet++);
	    if( clause_length[ clause_index ] == SAT_INCREASE )
	    {
		restore_big_occurences( clause_index, -nrval );
		clause_length[ clause_index ] = 2;
	    }

            clause_length[ clause_index ]++;
          }
	}

	/* restore all literals that were removed due to fixing of nrval */

	if( kSAT_flag == 0 )
	{
	    int *tImp = TernaryImp[ -nrval ];
	    for( i = TernaryImpSize[ -nrval ] = tmpTernaryImpSize[ -nrval ]; i--; )
	    {
	    	TernaryImpSize[ *(tImp++) ]++;
	    	TernaryImpSize[ *(tImp++) ]++;
	    }
	}

#ifdef GLOBAL_AUTARKY
        tImp = TernaryImp[ -nrval ];
        for( i = tmpTernaryImpSize[ -nrval ]; i--; )
	{
	    TernaryImpReduction[ *(tImp++) ]--;
	    TernaryImpReduction[ *(tImp++) ]--;
	}
#endif
	/* restore all clauses that were removed due to fixing of nrval */
	if( kSAT_flag == 0 )
	{
	    int *tImp = TernaryImp[ nrval ];
	    for( i = TernaryImpSize[ nrval ] = tmpTernaryImpSize[ nrval ]; i--; )
	    {
	    	TernaryImpSize[ *(tImp++) ]++;
	    	TernaryImpSize[ *(tImp++) ]++;
	    }
	}

#ifdef GLOBAL_AUTARKY
        tImp = TernaryImp[ nrval ] + 2 * TernaryImpSize[ nrval ];
        for( i = TernaryImpLast[ nrval ] - TernaryImpSize[ nrval ]; i--; )
        {
	    lit1 = *(tImp++);
	    lit2 = *(tImp++);

	    if( IS_REDUCED_TIMP( lit1, lit2 ) )
                TernaryImpReduction[ lit1 ]++;
	    else if( IS_REDUCED_TIMP( lit2, lit1 ) )
                TernaryImpReduction[ lit2 ]++;
        }
#endif
        bImp = BIMP_START(-nrval);
        for( i = BIMP_ELEMENTS; --i; )
            bImp_satisfied[ -(*(bImp++)) ]--;

	freevars++;

	UNFIX( nrval );
}

int checkSolution( )
{
	int i, *sizes;

	if( kSAT_flag ) sizes = big_occ;
	else 		sizes = TernaryImpSize;

	for ( i = 1; i <= original_nrofvars; i++ )
	    if( IS_NOT_FIXED( i ) )
	    {
	        if( sizes[  i ] > 0 ) { return UNSAT; }
	        if( sizes[ -i ] > 0 ) { return UNSAT; }
	        if( BinaryImp[  i ][ 0 ] > bImp_satisfied[  i ] ){ return UNSAT; }
	        if( BinaryImp[ -i ][ 0 ] > bImp_satisfied[ -i ] ){ return UNSAT; }
	    }

	return SAT;
}

//kan ook elders heen

int verifySolution()
{
	int i, j, satisfied;
#ifndef CUT_OFF
	unsigned int mask = 0xffffffff;

	if (DOOUTPUT) printf("\nc solution bin :: " );
	for( i = solution_bits; i >= 0; i-- )
	{
	    if( solution_bits - i == jump_depth && DOOUTPUT) printf(".");
	    if( ((solution_bin & mask) >> i) > 0 && DOOUTPUT) printf("1");
	    else if (DOOUTPUT) printf("0");
	    mask = mask >> 1;
	}
	if (DOOUTPUT)printf("\n");
#else
	if (DOOUTPUT) printf("s %i\n", solution_bin + 1);
	fflush( stdout );

        last_SAT_bin = solution_bin;

        return UNSAT;
#endif
#ifdef DISTRIBUTION
        if (DOOUTPUT)printf("c direction branchin :: current_rights = %i\n", current_rights );
#endif
#ifdef COUNT_SAT
	count_sat++;
	//printSolution( original_nrofvars );
	return UNSAT;
#endif
	do
	{ fixDependedEquivalences(); }
	while( dependantsExists() );

	/* check all 3-clauses */

	for( i = 0; i < nrofclauses; i++ )
	{
	    satisfied = 0;
	    if( Clength[ i ] == 0 ) continue;

	    for( j = 0; j < Clength[ i ]; j++ )
		if( timeAssignments[ Cv[ i ][ j ] ] == VARMAX ) satisfied = 1;

  	    if( !satisfied )
	    {
  	    	if (DOOUTPUT)printf("\nc clause %i: ", i);
		for( j = 0; j < Clength[ i ]; j++ )
			if (DOOUTPUT)printf("%i [%i] ", Cv[i][j], IS_FIXED(Cv[i][j]) );
		if (DOOUTPUT)printf("not satisfied yet\n");
		return UNKNOWN;
	    }
	}

#ifdef EQ
        for( i = 0; i < nrofceq; i++ )
        {
      	    int value = CeqValues[ i ];
            for( j = 0; j < CeqSizes[ i ]; j++ )
            	value *= EQSGN( Ceq[ i ][ j ] );
            if( value == -1 )
            {
            	if (DOOUTPUT)printf("c eq-clause %i is not satisfied yet\n", i);
            	return UNKNOWN;
            }
        }
#endif
	return SAT;
}

//kan ook elders heen...

void printSolution( const int orignrofvars , int marchResult[])
{
	int i;
#ifdef COMBINATORICS
	int j, pos = 0, neg = 0, rows;

	for( i = 1; i <= orignrofvars; i++ )
	    if     ( timeAssignments[ i ] ==  VARMAX      ) pos++;
	    else if( timeAssignments[ i ] == (VARMAX + 1) ) neg++;

	rows    = (neg + pos) / pos;

	if( rows > 1 && (neg % pos == 0) )
	{
	    for( j = 1; j <= rows; j++ )
	    {
	    	if (DOOUTPUT)printf("paritition %i :: ", j );
	    	for( i = j; i <= orignrofvars; i += rows )
	            if( timeAssignments[ i ] ==  VARMAX && DOOUTPUT) printf(" %i", (i+ rows -1) / rows );
	    	if (DOOUTPUT)printf("\n");
	    }
	    return;
	}
#endif
	if (DOOUTPUT)printf("c v");
	for( i = 1; i <= orignrofvars; i++ )
	    if     ( timeAssignments[ i ] ==  VARMAX      ) {
	    	if (DOOUTPUT)printf(" %i",  i);
	    	marchResult[i] = 1;
	    }
	    else if( timeAssignments[ i ] == (VARMAX + 1) ){
	    	if (DOOUTPUT)printf(" %i", -i);
	    	marchResult[i] = 0;
	    } else {
	    	marchResult[i] = 2;
	    }
	if (DOOUTPUT)printf(" 0\n");
}
